# -*- coding: utf-8 -*-

from asymmetree.cograph.Cograph import Cotree, CotreeNode
from asymmetree.cograph.LinearCographDetector import linear_cograph_detection, LinearCographDetector
from asymmetree.cograph.CographEditor import edit_to_cograph, CographEditor


# -*- coding: utf-8 -*-

"""
Assign asymmetric evolution rates to the branches of a gene tree.

Introduce evolution rate asymmetries and autocorrelation.
"""

import numpy as np

from asymmetree.treeevolve.GeneTree import GeneTreeSimulator


__author__ = 'David Schaller'

# --------------------------------------------------------------------------
#                         USER INTERFACE FUNCTION
# --------------------------------------------------------------------------

def simulate_gene_trees(S, N=1,
                        base_rate_distr=('constant', 1.0),
                        **kwargs):
    """Simulates dated gene trees with non-ultrametric edge lengths along a
    species tree.
    
    Keyword arguments:
        N -- number of gene trees to be simulated, default is 1, in which case
            a tree is returned, otherwise a list is returned
        base_rate_distr -- distribution for the evolution rate at the roots of
            the gene trees, default is ('constant', 1.0)
        kwargs -- see arguments of GeneTreeSimulator.simulate and assign_rates
        """
    
    gene_trees = []
    
    simulator = GeneTreeSimulator(S)
    
    # autocorrelation between genes of the same or related species
    autocorr_variance = kwargs.pop('autocorr_variance', 0.0)
    _, autocorr_factors = autocorrelation_factors(S, autocorr_variance)
    
    # main simulation and imbalancing
    for i in range(N):
        
        TGT = simulator.simulate(**kwargs)
        assign_rates(TGT, S,
                     base_rate=_get_base_rate(base_rate_distr),
                     autocorr_factors=autocorr_factors,
                     **kwargs)
        gene_trees.append(TGT)
    
    if N == 1:
        return gene_trees[0]
    else:
        return gene_trees
    

def _get_base_rate(distr):
    
    if isinstance(distr, (int, float)) and distr >= 0.0:
        return float(distr)
    
    elif isinstance(distr, (tuple, list)):
        
        if (distr[0] == 'constant' and
            isinstance(distr[1], (int, float)) and
            distr[1] >= 0.0):
            return float(distr[1])
        
        elif (distr[0] == 'gamma' and
              isinstance(distr[1], (int, float)) and distr[1] > 0.0 and
              isinstance(distr[2], (int, float)) and distr[2] > 0.0):
            shape = float(distr[1])
            scale = float(distr[2])
            return np.random.gamma(shape, scale=scale)
        
        elif (distr[0] == 'gamma_mean' and
              isinstance(distr[1], (int, float)) and distr[1] > 0.0):
            shape = 1.0
            scale = distr[1] / shape
            return np.random.gamma(shape, scale=scale)
        
    raise ValueError("distribution '{}' not supported".format(distr))
    

# --------------------------------------------------------------------------
#                      IMBALANCING OF THE GENE TREE
# --------------------------------------------------------------------------


def assign_rates(T, S, base_rate=1.0,
                 autocorr_factors=None,
                 autocorr_variance=0.0,
                 gamma_param=(0.5, 1.0, 2.2),
                 CSN_weights=(1, 1, 1),
                 inplace=True,
                 **kwargs):
    """Assigns realistic evolution rates to a TRUE gene tree.
    
    The assigned rates are used to modify the length ('dist') of the edges of
    the (originally ultrametric) dated gene tree.
    
    Keyword arguments:
    base_rate -- mean of substitution rate for conserved genes
    autocorr_factors -- autocorrelation rate factors for the edges of S
    autocorr_variance -- autocorrelation variance factor for lognormal
        distribution, only relevant if 'autocorrelation_rates' are not supplied 
    gamma_param -- param. for gamma distribution (a, loc, scale)
    CSN_weights -- weights for choice between conservation, subfunctionalization
        and neofunctionalization
    inplace -- if False, copy the tree before imbalancing
    """
    
    if not inplace:
        T = T.copy()
     
    # factors for subfunctionalization/neofunctionalization
    CSN_weights = np.asarray(CSN_weights) / sum(CSN_weights)
    _divergent_rates(T, S, gamma_param, CSN_weights)
    
    # autocorrelation
    if autocorr_factors:
        _apply_autocorrelation(T, autocorr_factors, inplace=True)
    elif autocorr_variance > 0.0:
        _, edge_rates = autocorrelation_factors(S, autocorr_variance)
        _apply_autocorrelation(T, edge_rates, inplace=True)
    
    # finally apply base rate
    for v in T.preorder():
        v.dist *= base_rate
    
    return T


# --------------------------------------------------------------------------
#                       EVOLUTION RATE ASYMMETRY
# --------------------------------------------------------------------------

def _adjust_distances(T, rates):
    
    for edge, rate_list in rates.items():
        time_points = np.asarray([tstamp for tstamp, _ in rate_list] + [edge[1].tstamp])
        rate_values = np.asarray([rate for _, rate in rate_list])
        edge[1].dist = np.dot(-np.diff(time_points), rate_values)
        

def _get_rate(marked_as, gamma_param):
    
    if marked_as == "conserved":
        # --- conserved genes are set to rate 1 ---
        return 1.0
    
    elif marked_as == "divergent":
        # --- fitted gamma distribution ---
        return (np.random.gamma(gamma_param[0], scale=gamma_param[2]) +
                gamma_param[1])
        

def _duplication_type(marked_as, CSN_weights):
    
    if marked_as == "divergent":
        return "divergent", "divergent"
    else:
        r = np.random.choice(3, p=CSN_weights)
        if r == 0:                                  # conservation
            return "conserved", "conserved"
        elif r == 1:                                # subfunctionalization
            return "divergent", "divergent"
        else:                                       # neofunctionalization
            if np.random.uniform() < 0.5:
                return "divergent", "conserved"
            else:
                return "conserved", "divergent"


def _divergent_rates(T, S, gamma_param, CSN_weights):
    """
    Keyword arguments:
    gamma_param -- param. for gamma distribution (a, loc, scale)
    CSN_weights -- weights for choice between conservation,
        subfunctionalization and neofunctionalization
    """
    
    T_nodes = T.sorted_nodes()
    rates = {edge: [] for edge in T.edges()}        # edge --> list of (tstamp, rate) tuples
    
    S_parents = {v.ID: v.parent.ID for v in S.preorder() if v.parent}
    gene_counter = {(e[0].ID, e[1].ID): [] for e in S.edges()}
    marked = {v: "conserved" for v in T_nodes}      # marked as conserved or divergent
    
    for u in T_nodes:
        
        # ----------------- SPECIATION -----------------
        if u.label in ("S", ""):
            for v in u.children:
                marked[v] = marked[u]
                S_u = u.color
                S_v = v.color if not isinstance(v.color, (tuple, list)) else v.color[1]
                gene_counter[(S_u, S_v)].append(v)
                rates[(u,v)].append((u.tstamp, _get_rate(marked[v], gamma_param)))
            
        # ---------------- DUPLICATION -----------------
        elif u.label == "D":
            marked[u.children[0]], marked[u.children[1]] = _duplication_type(marked[u],
                                                                             CSN_weights)
            gene_counter[u.color].remove(u)
            for v in u.children:
                gene_counter[u.color].append(v)
                rates[(u,v)].append((u.tstamp, _get_rate(marked[v], gamma_param)))
        
        # ------------------- LOSS ---------------------
        elif u.is_loss():
            gene_counter[u.color].remove(u)
            if len(gene_counter[u.color]) == 1:
                v = gene_counter[u.color][0]
                if marked[v] == "divergent":
                    marked[v] = "conserved"
                    rates[(v.parent,v)].append((u.tstamp, _get_rate(marked[v], gamma_param)))
        
        # ---------- HORIZONTAL GENE TRANSFER ----------
        elif u.label == "H":
            v1, v2 = u.children
            if v1.transferred:
                v1, v2 = v2, v1         # now v2 is the transferred copy
                
            # untransferred copy
            marked[v1] = marked[u]
            gene_counter[u.color].remove(u)
            gene_counter[u.color].append(v1)
            if u.parent:
                rates[(u,v1)].append((u.tstamp, rates[(u.parent,u)][-1][1]))
            else:
                rates[(u,v1)].append((u.tstamp, _get_rate(marked[v1], gamma_param)))
            
            # transferred copy
            marked[v2] = "divergent"
            if isinstance(v2.color, (tuple, list)):
                gene_counter[v2.color].append(v2)
            else:
                gene_counter[(S_parents[v2.color], v2.color)].append(v2)
            rates[(u,v2)].append((u.tstamp, _get_rate(marked[v2], gamma_param)))
            
    _adjust_distances(T, rates)
    return T

# --------------------------------------------------------------------------
#                         AUTOCORRELATION
# --------------------------------------------------------------------------
    
def autocorrelation_factors(tree, variance):
    """Geometric Brownian motion process to assign rate factors to species tree.
    
    The parameter 'variance' is a hyperparameter for a log-normal distribution
    from which offspring rates are drawn. The overall variance of this
    distribution is 'variance' * divergence time.
    """
    
    node_rates = {}                 # maps node v --> rate of v
    edge_rates = {}                 # maps v of edge (u,v) --> rate of (u,v)
    
    for v in tree.preorder():
        if not v.parent:
            # assign factor 1.0 to root (= expected value for all other nodes
            # and edges)
            node_rates[v.ID] = 1.0
            edge_rates[v.ID] = 1.0
        else:
            var = variance * v.dist
            # ensure that exp. value is equal to parent's rate
            mu = np.log(node_rates[v.parent.ID]) - var/2
            
            node_rates[v.ID] = np.exp(np.random.normal(mu, np.sqrt(var)))
            
            # edge rate as arithmetic mean of u and v
            edge_rates[v.ID] = (node_rates[v.parent.ID] + node_rates[v.ID]) / 2
            
    return node_rates, edge_rates


def _apply_autocorrelation(T, edge_rates, inplace=True):
    
    if not inplace:
        T = T.copy()
    
    for v in T.preorder():
        if v.parent:
            edge_ID = v.color[1] if isinstance(v.color, (tuple, list)) else v.color
            v.dist *= edge_rates[edge_ID]
    
    return T



# -*- coding: utf-8 -*-

"""
Orthology graph, cBMG, cRBMG.

This module provides classes concerning colored graphs and colored
(phylogentic) trees, including a cBMG(PhyloTree)-function.
"""

import itertools

import networkx as nx

from asymmetree.tools.GraphTools import symmetric_part


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                 True Orthology Graph (TOG), cBMG & cRBMG
#
#                         (given a known tree)
# --------------------------------------------------------------------------

def orthology_from_tree(tree):
    """Constructs the true orthology graph from an event-labeled tree."""
    
    tree.supply_leaves()                                # assign list of leaves to each node
    TOG = nx.Graph()
    
    for v in tree.root.leaves:
        TOG.add_node(v.ID, label=v.label, color=v.color)
    
    for node in tree.preorder():
        if node.label == 'S':
            for child1, child2 in itertools.combinations(node.children, 2):
                for u in child1.leaves:
                    for v in child2.leaves:
                        TOG.add_edge(u.ID, v.ID)
                        
    return TOG


def bmg_from_tree(tree, supply_rbmg=False):
    """Create an n-colored BMG (and optionally RBMG) from a given tree."""
    
    tree.supply_leaves()                                # assign list of leaves to each node
    bmg = nx.DiGraph()
    colors = set()
    
    for v in tree.root.leaves:
        colors.add(v.color)
        bmg.add_node(v.ID, label=v.label, color=v.color)
    
    for u in tree.root.leaves:
        remaining = colors - set([u.color])             # colors to which no best match has yet been found
        parent = u.parent                               # start with direct parent of each node
        while remaining and parent:
            colors_here = set()
            for v in parent.leaves:
                if v.color in remaining:                # best match found
                    colors_here.add(v.color)
                    bmg.add_edge(u.ID, v.ID)            # insert edge (u,v)
            remaining -= colors_here                    # update remaining colors
            parent = parent.parent
    
    if not supply_rbmg:
        return bmg
    else:
        return bmg, symmetric_part(bmg)


# -*- coding: utf-8 -*-

"""
Graph Tools.
"""

import itertools, random
import numpy as np
import networkx as nx


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                            GENERAL GRAPH TOOLS
# --------------------------------------------------------------------------

def graphs_equal(G1, G2):
    """Returns whether two NetworkX graphs (directed or undirected) are equal."""
    
    if (not G1.order() == G2.order()) or (not G1.size() == G2.size()):
        return False
    
    if set(G1.nodes()) != set(G2.nodes()):
        return False
    
    for u, v in G1.edges():
        if not G2.has_edge(u,v):
            return False
    
    return True


def symmetric_part(G):
    """Returns the (undirected) symmetric part of a directed graph."""
    
    if not isinstance(G, nx.DiGraph):
        raise TypeError('directed graph required')
    
    G_sym = nx.Graph()
    G_sym.add_nodes_from(G.nodes(data=True))
    
    for x, neighbors in G.adjacency():
        for y in neighbors:
            if G.has_edge(y,x):
                G_sym.add_edge(x,y)
                
    return G_sym


def symmetric_diff(G1, G2):
    """Returns the number of edges in the symmetric difference."""
    
    set1 = {x for x in G1.nodes()}
    set2 = {x for x in G2.nodes()}
    
    if set1 != set2:
        raise RuntimeError('graphs do not have the same vertex set')
        return
    
    sym_diff_number = 0
    
    for x, y in itertools.combinations(set1, 2):
        if G1.has_edge(x, y) and not G2.has_edge(x, y):
            sym_diff_number += 1
        elif not G1.has_edge(x, y) and G2.has_edge(x, y):
            sym_diff_number += 1
    
    return sym_diff_number


def random_graph(N, p=0.5):
    """Construct a random graph on N nodes.
    
    Keyword arguments:
        p - probability that an edge xy is inserted
    """
        
    G = nx.Graph()
    G.add_nodes_from(range(1, N+1))
    
    for x, y in itertools.combinations(range(1, N+1), 2):
        if random.random() < p:
            G.add_edge(x, y)
    
    return G


def performance(true_graph, graph):
    """Returns order, size, tp, tn, fp, fn, accuracy, precision and recall
    of a (directed or undirected) graph w.r.t. true graph."""
    
    order, size, tp, fp, fn = graph.order(), graph.size(), 0, 0, 0
    
    for u, v in graph.edges():
        if true_graph.has_edge(u,v):
            tp += 1
        else:
            fp += 1
    
    for u, v in true_graph.edges():
        if not graph.has_edge(u, v):
            fn += 1
    
    if isinstance(graph, nx.DiGraph):
        tn = (order * (order-1)) - (tp + fp + fn)
    else:
        tn = (order * (order-1) / 2) - (tp + fp + fn)
    
    accuracy = (tp + tn) / (tp + tn + fp + fn) if tp + tn + fp + fn > 0 else float('nan')
    precision = tp / (tp + fp) if tp + fp > 0 else float('nan')
    recall = tp / (tp + fn) if tp + fn > 0 else float('nan')
    
    return order, size, tp, tn, fp, fn, accuracy, precision, recall


def false_edges(true_graph, graph):
    """Returns a graph containing fn and a graph containg fp edges."""
    
    if isinstance(true_graph, nx.DiGraph):
        fn_graph = nx.DiGraph()
        fp_graph = nx.DiGraph()
    else:
        fn_graph = nx.Graph()
        fp_graph = nx.Graph()
        
    fn_graph.add_nodes_from(true_graph.nodes(data=True))
    fp_graph.add_nodes_from(true_graph.nodes(data=True))
        
    for u, v in graph.edges():
        if not true_graph.has_edge(u,v):
            fp_graph.add_edge(u, v)
    
    for u, v in true_graph.edges():
        if not graph.has_edge(u, v):
            fn_graph.add_edge(u, v)
        
    return fn_graph, fp_graph


def build_adjacency_matrix(G):
    """Return an adjacency matrix."""
    
    # maps node --> row/column index
    index = {node: i for i, node in enumerate(G.nodes())}
    M = np.zeros( (len(index), len(index)), dtype=np.int8)
    
    for x, neighbors in G.adjacency():
        for y in neighbors:
            M[index[x], index[y]] = 1
            
    return M


def sort_by_colors(graph):
    
    color_dict = {}
    
    for v in graph.nodes():
        color = graph.nodes[v]['color']
        
        if color not in color_dict:
            color_dict[color] = [v]
        else:
            color_dict[color].append(v)
    
    return color_dict
            

def classify_good_ugly(bmg, rbmg, fp):
    
    for x, y in fp.edges():
        fp.edges[x, y]['middle_in_good'] = False
        fp.edges[x, y]['first_in_ugly'] = False
        fp.edges[x, y]['first_in_bad'] = False
        
    color_dict = sort_by_colors(bmg)
    
    for x, y in fp.edges():
        
        # middle edges of good quartets
        for c in color_dict:
            
            if c == bmg.nodes[x]['color'] or c == bmg.nodes[y]['color']:
                continue
            
            for z1, z2 in itertools.permutations(color_dict[c], 2):
                
                if (rbmg.has_edge(z1, x) and rbmg.has_edge(z2, y) and
                    bmg.has_edge(z1, y) and not bmg.has_edge(y, z1) and
                    bmg.has_edge(z2, x) and not bmg.has_edge(x, z2)):
                    
                    fp.edges[x, y]['middle_in_good'] = True
        
        # first edges of ugly/bad quartets          
        for _ in range(2):
            
            for x2 in color_dict[bmg.nodes[x]['color']]:
                if x == x2:
                    continue
                
                for z in rbmg.neighbors(x2):
                    
                    if (bmg.nodes[z]['color'] == bmg.nodes[x]['color'] or
                        bmg.nodes[z]['color'] == bmg.nodes[y]['color']):
                        continue
                    # first edges of ugly quartets
                    if (rbmg.has_edge(y, x2) and not rbmg.has_edge(z, y) and
                        not rbmg.has_edge(z, x)):
                        fp.edges[x, y]['first_in_ugly'] = True
                    
                    # first edges of bad quartets
                    elif (not rbmg.has_edge(y, x2) and rbmg.has_edge(z, y) and
                          not bmg.has_edge(x, z)):
                        fp.edges[x, y]['first_in_bad'] = True
            
            # swap for second iteration
            x, y = y, x
    
    return fp   


def count_good_ugly(bmg, rbmg, fp):
    
    classify_good_ugly(bmg, rbmg, fp)
    
    good, ugly, both = 0, 0, 0
    
    for x, y in fp.edges():
        
        if fp.edges[x, y]['middle_in_good'] and fp.edges[x, y]['first_in_ugly']:
            good += 1
            ugly += 1
            both += 1
        elif fp.edges[x, y]['middle_in_good']:
            good += 1
        elif fp.edges[x, y]['first_in_ugly']:
            ugly += 1
    
    return good, ugly, both


def count_good_ugly_bad(bmg, rbmg, fp):
    
    classify_good_ugly(bmg, rbmg, fp)
    
    gub, gu, gb, ub, g, u, b = 0, 0, 0, 0, 0, 0, 0
    
    for x, y in fp.edges():
        
        if (fp.edges[x, y]['middle_in_good'] and
            fp.edges[x, y]['first_in_ugly'] and
            fp.edges[x, y]['first_in_bad']):
            gub += 1
            gu += 1
            gb += 1
            ub += 1
            g += 1
            u += 1
            b += 1
            
        elif fp.edges[x, y]['middle_in_good'] and fp.edges[x, y]['first_in_ugly']:
            gu += 1
            g += 1
            u += 1
        elif fp.edges[x, y]['middle_in_good'] and fp.edges[x, y]['first_in_bad']:
            gb += 1
            g += 1
            b += 1
        elif fp.edges[x, y]['first_in_ugly'] and fp.edges[x, y]['first_in_bad']:
            ub += 1
            u += 1
            b += 1
        elif fp.edges[x, y]['middle_in_good']:
            g += 1
        elif fp.edges[x, y]['first_in_ugly']:
            u += 1
        elif fp.edges[x, y]['first_in_bad']:
            b += 1
    
    return gub, gu, gb, ub, g, u, b
    

# --------------------------------------------------------------------------
#                             P4 EDITING
# --------------------------------------------------------------------------
    
def list_path(t, k, path, G, P4_list):
    for u in G[path[-1]]:
        G.nodes[u]['vis'] += 1
    for u in G[path[-1]]:
        if G.nodes[u]['vis'] == 1:
            if t < k:
                list_path(t+1, k, path + [u], G, P4_list)
            elif path[0] < u:
                P4_list.append(tuple(path) + (u,))
    for u in G[path[-1]]:
        G.nodes[u]['vis'] -= 1


def find_all_P4(G):
    P4_list = []
    for v in G:
            G.nodes[v]['vis'] = 0
    for v in G:
        G.nodes[v]['vis'] = 1
        list_path(2, 4, [v], G, P4_list)
        G.nodes[v]['vis'] = 0
    return P4_list


def is_good_quartet(path, bmg):
    """Determine whether an induced P4 is a good quartet in the bmg."""
    if (bmg.has_edge(path[0], path[2]) and 
        bmg.has_edge(path[3], path[1]) and
        bmg.nodes[path[0]]['color'] == bmg.nodes[path[3]]['color']):
        return True
    else:
        return False
    

def remove_all_P4(rbmg, bmg, P4_list=None):
    """Find all good quartets and removes the middle edges."""
    GP4 = rbmg.copy()
    
    if P4_list is None:
        P4_list = find_all_P4(GP4)

    for path in P4_list:
        if is_good_quartet(path, bmg) and GP4.has_edge(path[1], path[2]):
            GP4.remove_edge(path[1], path[2])
            print(path[1], "-|-", path[2])
    
    return GP4

# --------------------------------------------------------------------------
#                            C6 DETECTION
# --------------------------------------------------------------------------

def validate_C6(G, C6_path):
    
    C6 = G.subgraph(C6_path)
    if C6.size() != 6:
        print(C6_path)
        print('Nodes', [n for n in C6.nodes()], C6.order())
        print('Edges', [edge for edge in C6.edges()], C6.size())
        print("Size not equal to 6!")
        return False
    for n in C6.nodes():
        if not C6.degree(n) == 2:
            print("Degree not equal to 2 for", n)
            return False
        for neighbor in G.neighbors(n):
            if G.nodes[n]['color'] == G.nodes[neighbor]['color']:
                print("Same color", n, neighbor)
                return False
    return True


def find_all_C6(G, P4_list=None):
    """Find induces C6 of the form <x1 y1 z1 x2 y2 z2>."""
    
    if P4_list is None:
        P4_list = find_all_P4(G)
        
    endpoints = {}
    for path in P4_list:
        a, b, c, d = path if path[0] < path[3] else path[::-1]
        a_col, b_col, c_col, d_col = (G.nodes[a]['color'], G.nodes[b]['color'],
                                      G.nodes[c]['color'], G.nodes[d]['color'])
        # avoid multiple recording for the 3 different colors:
        if a_col == d_col and a_col < b_col and a_col < c_col:
            if (a, d) not in endpoints:
                endpoints[(a, d)] = {}
            if (b_col, c_col) not in endpoints[(a, d)]:
                endpoints[(a, d)][(b_col, c_col)] = []
            endpoints[(a, d)][(b_col, c_col)].append(path)
    
    C6_list = []
    for endpoint, colors in endpoints.items():
        x1, x2 = endpoint
        for color_pair, paths in colors.items():
            y_color, z_color = color_pair
            # avoid multiple recording for the 2 directions:
            if y_color < z_color and (z_color, y_color) in colors:
                for path1 in paths:
                    for path2 in colors[(z_color, y_color)]:
                        if ((not G.has_edge(path1[1], path2[1])) and
                            (not G.has_edge(path1[2], path2[2]))):
                            C6_list.append(path1 + (path2[2], path2[1]))
                            if not validate_C6(G, C6_list[-1]):
                                raise RuntimeError('problem in C6 detection')
                            print(C6_list[-1])
    return C6_list


def graph_type(G):
    
    P4_list = find_all_P4(G)
    C6_list = find_all_C6(G, P4_list=P4_list)
    
    if C6_list:
        return 'C', P4_list, C6_list
    elif P4_list:
        return 'B', P4_list, C6_list
    else:
        return 'A', P4_list, C6_list



# -*- coding: utf-8 -*-

from asymmetree.seqevolve.Evolver import Evolver
from asymmetree.seqevolve.SubstModel import SubstModel
from asymmetree.seqevolve.IndelModel import IndelModel
from asymmetree.seqevolve.HetModel import HetModel
from asymmetree.seqevolve.Alignment import AlignmentBuilder
from asymmetree.seqevolve.EvolvingSequence import EvoSeq


# -*- coding: utf-8 -*-

import numpy as np

from asymmetree.seqevolve.EvolvingSequence import EvoSeq, State
from asymmetree.seqevolve.Alignment import AlignmentBuilder
from asymmetree.file_io.SeqFileIO import write_alignment


__author__ = 'David Schaller'


class Evolver:
    
    def __init__(self, subst_model,
                 indel_model=None, het_model=None,
                 jump_chain=False,
                 **kwargs):
        
        self.subst_model = subst_model
        
        self.indel_model = indel_model
        self.het_model = het_model
        
        if (jump_chain or 
            (het_model and het_model.sitewise) or
            (het_model and het_model.classes > 5)):     # determine best cutoff
            self._jump_chain = True
        else:
            self._jump_chain = False
        
    
    def evolve_along_tree(self, tree, start_length=200, start_seq=None):
        
        self.tree = tree
        self.site_counter = 0
        self.sequences = {}
        
        if start_seq is None:
            root_seq = self._random_sequence(start_length)
        else:
            root_seq = self._initialize_with_sequence(start_seq)
            
        if self.het_model:
            self.het_model.assign(root_seq)
        
        for v in tree.preorder():
            
            if v.parent is None:
                self.sequences[v] = root_seq
            else:
                self.sequences[v] = self._evolve(self.sequences[v.parent], v.dist)
                
        return self.sequences
          
            
    def _random_positions(self, n):
        
        return np.random.choice(len(self.subst_model.alphabet), n,
                                p=self.subst_model.freqs)
    
    
    def _random_sequence(self, n):
        
        seq = EvoSeq()
        
        for x in self._random_positions(n):
            seq.append(x, State.ROOT, self.site_counter)
            self.site_counter += 1
        
        return seq
    
    
    def _initialize_with_sequence(self, sequence):
        
        seq = EvoSeq()
        
        for x in self.subst_model.to_indices(sequence):
            seq.append(x, State.ROOT, self.site_counter)
            self.site_counter += 1
        
        return seq
    
    
    def _evolve(self, parent_seq, distance):
        
        child_seq = parent_seq.clone()
        
        if self.indel_model:
            self._generate_indels(child_seq, distance)
            
            if self.het_model:
                self.het_model.assign(child_seq)
        
        if not self._jump_chain:
            self._substitute(child_seq, distance)               # apply transition prob. matrix
        else:
            self._substitute_jump_chain(child_seq, distance)    # apply jump chain
        
        return child_seq
    
    
    def true_alignment(self, include_inner=True, write_to=None,
                       alignment_format='phylip'):
        
        alg_builder = AlignmentBuilder(self.tree, self.sequences,
                                       self.subst_model.alphabet,
                                       include_inner=include_inner)
        
        alignment = alg_builder.build()
        
        if write_to:
            write_alignment(write_to, alignment,
                            alignment_format=alignment_format)
        
        return alignment
    
    
    # --------------------------------------------------------------------------
    #               Substitution by transition probability matrix
    # --------------------------------------------------------------------------
    
    def _substitute(self, sequence, t):
        
        # (cumulative) transition probability matrices
        P = self._cumulative_P_matrices(sequence, t)
        
        r = np.random.random(len(sequence))
        pos = 0
        
        for site in sequence:
            
            if site.status == State.INHERITED:
                # mutate according to matrix of the corresponding class
                site._value = np.argmax(P[site.rate_class][site._value, :] > r[pos])
            
            else:
                # choose random character
                site._value = np.argmax(self.subst_model.freqs_cumulative > r[pos])
            
            pos += 1
    
    
    def _cumulative_P_matrices(self, sequence, t):
        
        P = {}
        
        if not self.het_model:
            P[0] = np.cumsum(self.subst_model.transition_prob_matrix(t),
                             axis=1)

        else:
            for site in sequence:
                
                c = site.rate_class
                
                if c not in P:
                    
                    r = site.rate_factor
                    
                    if r > 0.0:
                        P[c] = np.cumsum(self.subst_model.transition_prob_matrix(r * t),
                                         axis=1)
                    elif r == 0:
                        P[c] = np.cumsum(np.identity(len(self.subst_model.alphabet)),
                                         axis=1)
        
        return P
        
    
    # --------------------------------------------------------------------------
    #                        Substitution by Jump chain
    # --------------------------------------------------------------------------
    
    def _substitute_jump_chain(self, sequence, t):
        
        total_rate = self._total_subst_rate(sequence, exclude_inserted=True)
        
        # jump chain algorithm for inherited positions
        current_time = 0.0
        while current_time < t:
            
            current_time += np.random.exponential(1/total_rate) if total_rate > 0.0 else float('inf')
            
            if current_time < t:
                
                site, mutation = self._draw_substitution(sequence, total_rate)
                
                total_rate += site.rate_factor * self.subst_model.Q[site._value, site._value]
                total_rate -= site.rate_factor * self.subst_model.Q[mutation, mutation]
                site._value = mutation
        
        # choose random characters for insertions
        r = np.random.random(sequence.count_status(State.INSERTION))
        pos = 0
        for site in sequence:
            
            if site.status == State.INSERTION:
                site._value = np.argmax(self.subst_model.freqs_cumulative > r[pos])
                pos += 1
                
    
    def _total_subst_rate(self, sequence, exclude_inserted=True):
        
        total = 0.0
        
        for site in sequence:
            
            if exclude_inserted and site.status == State.INSERTION:
                continue
            
            total -= site.rate_factor * self.subst_model.Q[site._value, site._value]
            
        return total
    
    
    def _draw_substitution(self, sequence, total_rate):
        
        chosen_site, chosen_mutation = None, None
        r = np.random.uniform(low=0.0, high=total_rate)
        
        current_sum = 0.0
        for site in sequence:
            
            if site.status == State.INSERTION:
                continue
            
            # use negative value on the diagonal of the rate matrix Q
            site_rate = - site.rate_factor * self.subst_model.Q[site._value, site._value]
            current_sum += site_rate
            
            if current_sum > r:
                chosen_site = site
                
                # rescale the "rest" of r for application on the rate matrix Q
                r = (r - current_sum + site_rate) / site.rate_factor
                
                break
            
        current_sum = 0.0
        for i in range(len(self.subst_model.alphabet)):
            
            if i != site._value:            # skip the entry on diagonal
                current_sum += self.subst_model.Q[site._value, i]
                
                if current_sum > r:
                    chosen_mutation = i
                    break
                    
        return chosen_site, chosen_mutation
    
    
    # --------------------------------------------------------------------------
    #                               Indels
    # --------------------------------------------------------------------------
        
    def _generate_indels(self, sequence, t):
        """Generates indels by a Gillespie process."""
        
        current_time = 0.0
        
        while current_time < t:
            
            ins_rate, del_rate = self.indel_model.get_rates(len(sequence))
            total_rate = ins_rate + del_rate
            
            current_time += np.random.exponential(1/total_rate) if total_rate > 0.0 else float('inf')
            
            if current_time < t:
                
                r = np.random.random()
                if r < ins_rate / total_rate:
                    self._insertion(sequence)
                else:
                    self._deletion(sequence)
    
    
    def _insertion(self, sequence):
        
        d = self.indel_model.draw_length()
        pos = np.random.randint(-1, high=len(sequence))
        
        # initialize insertion before the first item
        if pos == -1:
            current_site = sequence.append_left(None, State.INSERTION,
                                                self.site_counter)
            self.site_counter += 1
            d -= 1
            
        # go to the element at position pos
        else:
            current_site = sequence.element_at(pos)
            
        for _ in range(d):
            current_site = sequence.insert_right_of(current_site,
                                                    None, State.INSERTION,
                                                    self.site_counter)
            
            self.site_counter += 1
    
    
    def _deletion(self, sequence):
        
        d = self.indel_model.draw_length()
        pos = np.random.randint(-d + 1, high=len(sequence))
        
        # deletion begins before or at the start of the sequence
        if pos <= 0:
            d += pos
            pos = 0
            
        sequence.remove_range(pos, d)


# -*- coding: utf-8 -*-

"""
Doubly-linked list.

Implementation of a doubly-linked list in Python 3. The list enables access
to single list elements in order to modify/delete values in constant time.
"""

import collections


__author__ = 'David Schaller'


class DLListElement:
    """Doubly-linked list element."""
    
    __slots__ = ('_value', '_prev', '_next')
    
    def __init__(self, value, prev_el=None, next_el=None):
        
        self._value = value
        self._prev = prev_el
        self._next = next_el


class DLList:
    """Doubly-linked list."""
    
    __slots__ = ('_first', '_last', '_count')
    
    def __init__(self, *args):
        
        self._first = None
        self._last = None
        self._count = 0
        for arg in args:
            if isinstance(arg, collections.abc.Iterable):
                for item in arg:
                    self.append(item)
            else:
                self.append(item)
        
    
    def __len__(self):
        
        return self._count
    
    
    def __nonzero__(self):
        
        return True if self._count > 0 else False
    
    
    def __iter__(self):
        
        return DLListIterator(self)


    def __next__(self):
        
        pass
    
    
    def __getitem__(self, index):
        
        return self.element_at(index)._value
    
    
    def first(self):
        
        return self._first._value
    
    
    def last(self):
        
        return self._last._value
    
    
    def first_element(self):
        
        return self._first
    
    
    def element_at(self, index):
        
        if not isinstance(index, int):
            raise TypeError("index must be of type 'int'")
            
        if index < (-self._count) or index >= self._count:
            raise IndexError('index {} is out of bounds'.format(index))
            
        if index < 0:
            index += self._count
        
        if index <= self._count // 2:
            # start from beginning
            element = self._first
            for _ in range(index):
                element = element._next
        
        else:
            # start from end
            element = self._last
            for _ in range(self._count - index - 1):
                element = element._prev
            
        return element
    
    
    def append(self, value):
        
        new_end = DLListElement(value, prev_el=self._last)
        if self._last:
            self._last._next = new_end
        self._last = new_end
        if not self._first:
            self._first = new_end
        self._count += 1
        return new_end
    
    
    def extend(self, iterable):
        
        for value in iterable:
            self.append(value)
    
    
    def append_left(self, value):
        
        new_start = DLListElement(value, next_el=self._first)
        if self._first:
            self._first._prev = new_start
        self._first = new_start
        if not self._last:
            self._last = new_start
        self._count += 1
        return new_start
    
    
    def remove_element(self, element):
        """Remove an item by reference to the 'DLListElement' instance in O(1)."""
        
        if element._prev:
            element._prev._next = element._next
        if element._next:
            element._next._prev = element._prev
        if self._first is element:
            self._first = element._next
        if self._last is element:
            self._last = element._prev
        element._prev, element._next = None, None
        self._count -= 1
    
    
    def remove(self, value):
        """Remove an item by value in O(n)."""
        
        element = self._first
        while element:
            if element._value == value:
                self.remove_element(element)
                return
            element = element._next
        raise KeyError('value {} is not in the doubly-linked list'.format(value))
        
        
    def remove_range(self, index, length=None):
        """Removes a range from the index of the specified length.
        
        Removes the range [index, index+length) from the sequence. If no length
        is specified or index+length is out of bounds, the list gets truncated."""
        
        if not isinstance(index, int):
            raise TypeError("index must be of type 'int'")
        elif index < 0:
            index = self._count + index
            
        if length is not None and (not isinstance(length, int) or length < 1):
            raise TypeError("length must be of type 'int' and >0")
        
        if length is None or index + length >= self._count:
            self.truncate(index)
            
        elif index == 0:
            self.truncate_left(length)
            
        else:
            cut_start = self.element_at(index)
            cut_end = self.element_at(index + length - 1)
            
            cut_start._prev._next = cut_end._next
            cut_end._next._prev = cut_start._prev
            cut_start._prev, cut_end._next = None, None
            
            self._count -= length
    
    
    def insert_right_of(self, element, value):
        """Insert a new item right of an element of the list in O(1)."""
        
        if element is self._last:
            new_element = self.append(value)
            
        else:
            new_element = DLListElement(value,
                                        prev_el=element,
                                        next_el=element._next)
            new_element._next._prev = new_element
            element._next = new_element
            self._count += 1
        
        return new_element
            
    
    def truncate(self, index):
        """Truncate all elements starting from the specified index."""
        
        if index <= 0:
            self.clear()
        else:
            new_end = self.element_at(index-1)
            self._last = new_end
            if new_end._next:
                new_end._next._prev = None
                new_end._next = None
            self._count = index
            
    
    def truncate_left(self, n):
        """Truncate n elements on the left side."""
        
        if n >= self._count:
            self.clear()
        else:
            new_start = self.element_at(n)
            self._first = new_start
            if new_start._prev:
                new_start._prev._next = None
                new_start._prev = None
            self._count -= n
        
    
    def popright(self):
        """Removes the last element of the list and returns its value."""
        
        if self._last:
            value = self._last._value
            self.remove_element(self._last)
            return value
        else:
            return None
    
    
    def popleft(self):
        """Removes the first element of the list and returns its value."""
        
        if self._first:
            value = self._first._value
            self.remove_element(self._first)
            return value
        else:
            return None
    
    
    def clear(self):
        
        self._first = None
        self._last = None
        self._count = 0
        
    
    def _count_actual(self):
        """Counts the actual number of elements."""
        
        current = self._first
        counter = 0
        
        while current:
            counter += 1
            current = current._next
            
        return counter
        

class DLListIterator:
    """Iterator class for doubly-linked list."""
    
    def __init__(self, dllist):
        
        self.dllist = dllist
        self._current = dllist._first
        
    
    def __next__(self):
        
        if self._current:
            x = self._current
            self._current = self._current._next
            return x._value
        else:
            raise StopIteration


# -*- coding: utf-8 -*-


import networkx as nx

from asymmetree import TreeNode, PhyloTreeNode


__author__ = 'David Schaller'


def augment_and_label(tree, inplace=False):
    """Augment tree and add event labeling based on color intersections."""
    
    if not inplace:
        tree = tree.copy()
    
    # precompute since tree will be modified
    inner_vertices = [u for u in tree.inner_vertices()]
    tree.supply_leaves()
    current_ID = tree.get_max_ID() + 1
    
    for u in inner_vertices:
        
        # should not appear except if tree is planted, otherwise ignore
        if len(u.children) == 1:
            continue
        
        C = _color_intersection_components(u)
        
        if len(C) == 1:
            u.label = 'D'
            
        else:
            u.label = 'S'
            
            for cc in C:
                if len(cc) > 1:
                    w = PhyloTreeNode(current_ID, label='D')
                    current_ID += 1
                    u.add_child(w)
                    for v in cc:
                        u.remove_child(v)
                        w.add_child(v)
        
    return tree


def _color_intersection_components(u):
    
    aux_graph = nx.Graph()
    
    for v in u.children:
        aux_graph.add_node(v)
        
        for x in v.leaves:
            aux_graph.add_edge(v, x.color)
            
    result = []
    
    for cc in nx.connected_components(aux_graph):
        
        # discard the colors in the components
        result.append( [v for v in cc if isinstance(v, TreeNode)] )
        
    return result


if __name__ == '__main__':
    
    from asymmetree import PhyloTree
    from asymmetree.best_matches.LRTConstructor import lrt_from_observable_tree
    
    T = PhyloTree.random_colored_tree(30, 5)
    print('--- T ---\n', T.to_newick())
    
    lrt = lrt_from_observable_tree(T)
    print('--- LRT ---\n', lrt.to_newick())
    
    aug_tree = augment_and_label(lrt)
    print('--- Augmented tree ---\n', aug_tree.to_newick())


# -*- coding: utf-8 -*-

import itertools, random
import networkx as nx

from asymmetree.tools import Tree


__author__ = 'David Schaller'
    

class CotreeNode(Tree.TreeNode):
    
    __slots__ = ['aux_counter']
    
    
    def __init__(self, ID, label=None, parent=None,
                 aux_counter=0):
        
        super().__init__(ID, label=label)
        
        self.aux_counter = aux_counter      # auxiliary counting variable
        
    
    def __str__(self):
        
        if not self.children:
            return str(self.ID)
        
        elif self.label == 'series':
            return '<1>'
        
        elif self.label == 'parallel':
            return '<0>'
        
        else:
            return '<>'
        

class Cotree(Tree.Tree):
    
    # corresponding node type
    node_type = CotreeNode
    
    
    def __init__(self, root):
        super().__init__(root)
        
    
    def to_cograph(self):
        
        self.supply_leaves()
        G = nx.Graph()
        
        for v in self.root.leaves:
            G.add_node(v.ID)
        
        for u in self.preorder():
            if u.label == 'series':
                for v1, v2 in itertools.combinations(u.children, 2):
                    for l1, l2 in itertools.product(v1.leaves, v2.leaves):
                        G.add_edge(l1.ID, l2.ID)
        
        return G
        
        
    @staticmethod
    def cotree(G):
        """Checks if a graph is a cograph and returns its cotree.
        
        Simple O(n^3) implementation.
        """
        
        def build_cotree(G, label=None):
            
            v = CotreeNode(None)
            v.label = label
            child_nodes = []
            
            if G.order() == 1:
                v.label = 'leaf'
                for ID in G.nodes():
                    v.ID = ID
                    
                return v
            
            child_nodes = []
            
            if v.label is None:
                ccs = [cc for cc in nx.connected_components(G)]
                
                if len(ccs) > 1:
                    v.label = 'parallel'
                else:
                    G = nx.complement(G)
                    ccs = [cc for cc in nx.connected_components(G)]
                    if len(ccs) > 1:
                        v.label = 'series'
                    else:
                        return False
            
            else:
                G = nx.complement(G)
                ccs = [cc for cc in nx.connected_components(G)]
                
                if len(ccs) == 1:
                    return False
            
            child_label = 'series' if v.label == 'parallel' else 'parallel'
            for cc in ccs:
                G_i = G.subgraph(cc).copy()
                v_i = build_cotree(G_i, label=child_label)
                if v_i:
                    child_nodes.append(v_i)
                else:
                    return False
            
            for child in child_nodes:
                v.children.append(child)
                child.parent = v
            
            return v
        
        root = build_cotree(G)
        
        if root:
            return Cotree(root)
        else:
            return False
       
    
    @staticmethod
    def random_cotree(N_leaves, force_series_root=False):
        """Creates a random cotree."""
        
        root = CotreeNode(ID=0)
        cotree = Cotree(root)
        node_list = [root]
        nr, leaf_count = 1, 1
        
        while leaf_count < N_leaves:
            node = random.choice(node_list)
            
            # avoid nodes with outdegree 1
            if not node.children:
                new_child1 = CotreeNode(nr)
                new_child2 = CotreeNode(nr+1)
                node.add_child(new_child1)
                node.add_child(new_child2)
                node_list.extend(node.children)
                nr += 2
                leaf_count += 1
                
            # add only one child if there are already children
            elif node.children:
                new_child = CotreeNode(nr)
                node.add_child(new_child)
                node_list.append(new_child)
                nr += 1
                leaf_count += 1
        
        # assign labels ('series', 'parallel', 'leaf')
        for v in cotree.preorder():
            if not v.children:
                v.label == 'leaf'
            elif v.parent is None:
                if force_series_root:
                    v.label = 'series'
                else:
                    v.label = 'series' if random.random() < 0.5 else 'parallel'
            else:
                v.label = 'series' if v.parent.label == 'parallel' else 'parallel'
                
        return cotree


# -*- coding: utf-8 -*-

from asymmetree.treeevolve.GeneTree import simulate_dated_gene_tree, GeneTreeSimulator, observable_tree
from asymmetree.treeevolve.SpeciesTree import simulate_species_tree, simulate_species_tree_age
from asymmetree.treeevolve.EvolutionRates import assign_rates, autocorrelation_factors, simulate_gene_trees
from asymmetree.treeevolve.Scenario import Scenario
from asymmetree.treeevolve.NoisyMatrix import noisy_matrix, convex_linear_comb, wrong_topology_matrix


# -*- coding: utf-8 -*-

import itertools, random

import asymmetree.tools.DoublyLinkedList as dll


__author__ = 'David Schaller'


class TreeNode:
    """Class 'TreeNode'.
    
    Components for class 'Tree'. Contains a list of children as well as a
    reference to the parent node.
    """
    
    __slots__ = ('ID', 'label', 'parent', 'parent_dll_element',
                 'children', 'leaves')
    
    
    def __init__(self, ID, label=""):
        
        self.ID = ID
        self.label = label
        self.parent = None
        self.children = dll.DLList()
        self.parent_dll_element = None      # reference to doubly-linked list element
                                            # in the parents' children
    
    def __str__(self):
        
        return str(self.label)
    
    
    def __repr__(self):
        
        return "<TN: {}>".format(self.ID)
    
                                            
    def add_child(self, child_node):
        
        # do nothing if child_node is already a child of self
        
        if child_node.parent is None:
            child_node.parent = self
            child_node.parent_dll_element = self.children.append(child_node)
        
        elif child_node.parent is not self:
            child_node.parent.remove_child(child_node)
            child_node.parent = self
            child_node.parent_dll_element = self.children.append(child_node)  
    
    
    def remove_child(self, child_node):
        
        if child_node.parent is self:
            self.children.remove_element(child_node.parent_dll_element)
            child_node.parent = None
            child_node.parent_dll_element = None
        else:
            raise ValueError("{} is not a child of node {}".format(child_node, self))
            
            
    def detach(self):
        
        if self.parent is not None:
            self.parent.remove_child(self)
        else:
            self.parent = None
            self.parent_dll_element = None
        

class Tree:
    
    # corresponding node type
    node_type = TreeNode
    
    
    def __init__(self, root):
        
        self.root = root
    
    
    def preorder(self):
        """Generator for preorder traversal of the tree."""
        
        def _preorder(node):
            yield node
            for child in node.children:
                yield from _preorder(child)
        
        if self.root:
            yield from _preorder(self.root)
        else:
            yield from []
    
    
    def postorder(self):
        """Generator for postorder traversal of the tree."""
        
        def _postorder(node):
            for child in node.children:
                yield from _postorder(child)
            yield node
        
        if self.root:
            yield from _postorder(self.root)
        else:
            yield from []
            
    
    def inner_vertices(self):
        """Generator for inner vertices in preorder."""
        
        def _inner_vertices(node):
            if node.children:
                yield node
                for child in node.children:
                    yield from _inner_vertices(child)
        
        if self.root:
            yield from _inner_vertices(self.root)
        else:
            yield from []
            
    
    def edges(self):
        """Generator for all edges of the tree."""
        
        def _edges(node):
            for child in node.children:
                yield (node, child)
                yield from _edges(child)
        
        if self.root:
            yield from _edges(self.root)
        else:
            yield from []
            
            
    def edges_sibling_order(self):
        """Generator for all edges of the tree with sibling order.
        
        Returns edges uv as tuples (u, v, nr) where nr is the index of v in the
        list of children of node u."""
        
        def _edges_sibling_order(node):
            i = 0
            for child in node.children:
                yield (node, child, i)
                yield from _edges_sibling_order(child)
                i += 1
        
        if self.root:
            yield from _edges_sibling_order(self.root)
        else:
            yield from []
        
    
    def inner_edges(self):
        """Generator for all inner edges of the tree."""
        
        def _inner_edges(node):
            for child in node.children:
                if child.children:
                    yield (node, child)
                    yield from _inner_edges(child)
        
        if self.root:
            yield from _inner_edges(self.root)
        else:
            yield from []
                
    
    def euler_generator(self, id_only=False):
        """Generator for an Euler tour of the tree."""
        
        def _euler_generator(node, id_only):
            if id_only: yield node.ID
            else:       yield node
            
            for child in node.children:
                yield from _euler_generator(child, id_only)
                
                if id_only: yield node.ID
                else:       yield node
        
        if self.root:
            yield from _euler_generator(self.root, id_only)
        else:
            yield from []
            
    
    def supply_leaves(self):
        """Add the leaves to all nodes that are in the subtree of a specific node."""
        
        def _supply_leaves(node):
            node.leaves = []
            
            if not node.children:
                node.leaves.append(node)
            else:
                for child in node.children:
                    node.leaves.extend(_supply_leaves(child))
                    
            return node.leaves
        
        if self.root:
            return _supply_leaves(self.root)
        else:
            return []
        
        
    def get_triples(self):
        """Retrieve a list of all triples of the tree.
        
        Warning: Algorithm works recursively and can produce extremely 
        large lists.
        """
                
        self.supply_leaves()
        triples = []
        
        for u in self.preorder():
            for v1, v2 in itertools.permutations(u.children, 2):
                if len(v2.leaves) > 1:
                    for t3 in v1.leaves:
                        for t1, t2 in itertools.combinations(v2.leaves, 2):
                            triples.append( (t1, t2, t3) )
        
        return triples
    
    
    def delete_and_reconnect(self, node):
        """Delete a node from the tree and reconnect its parent and children."""
        
        parent = node.parent
        if not parent:
            print("Cannot delete and reconnect root '{}'!".format(node))
            return False
        else:
            parent.remove_child(node)
            
            # copy list of children to edit edges
            children = [child for child in node.children]
            for child in children:
                parent.add_child(child)
                    
            node.children.clear()
        
        return parent
    
    
    def get_max_ID(self):
        """Returns the maximum of all node IDs."""
        
        max_ID = -1
        
        for v in self.preorder():
            if v.ID is not None:
                max_ID = max(v.ID, max_ID)
                
        return max_ID
        
    
    def to_newick(self, node=None):
        """Recursive Tree --> Newick (str) function."""
        
        def _to_newick(node):
            if not node.children:
                return str(node)
            else:
                s = ''
                for child in node.children:
                    s += _to_newick(child) + ','
                return "({}){}".format(s[:-1], node)
        
        if self.root:
            return _to_newick(self.root) + ';'
        else:
            return ';'
      
    
    def get_hierarchy(self):
        """Return the hierarchy set on the leaf labels defined by the tree."""
        
        self.supply_leaves()
        
        hierarchy = set()
        
        for v in self.preorder():
            
            A = [leaf.label for leaf in v.leaves]
            A.sort()
            A = tuple(A)
            hierarchy.add(A)
            
        return hierarchy
    
    
    def compare_topology(self, other):
        """Compare the tree topology based on the hierarchies.
        
        Only works for binary trees."""
        
        hierarchy1 = sorted(self.get_hierarchy())
        hierarchy2 = sorted(other.get_hierarchy())
        
        if len(hierarchy1) != len(hierarchy2):
            print("Unequal sizes of the hierarchy sets: {} and {}".format(len(hierarchy1), len(hierarchy2)))
            return False
        
        for i in range(len(hierarchy1)):
            
            if hierarchy1[i] != hierarchy2[i]:
                print("Hierarchies not equal:\n{}\n{}".format(hierarchy1[i], hierarchy2[i]))
                return False
        
        return True
    
    
    def _assert_integrity(self):
        
        for v in self.preorder():
            for child in v.children:
                if child is v:
                    raise RuntimeError('loop at {}'.format(v))
                if child.parent is not v:
                    raise RuntimeError('Tree invalid for '\
                                       '{} and {}'.format(v, child))
        
        return True
    
    
    @staticmethod
    def random_tree(N, binary=False):
        """Creates a random tree.
        
        The number of leaves is specified by the parameter 'N'. Each non-leaf
        node in the resulting tree will have at least children (property of
        phylogenetic trees).
        
        Keyword arguments:
            binary - forces the tree to be binary; default is False
        """
        
        if not (isinstance(N, int)) or N < 1:
            raise TypeError("N must be an 'int' > 0")
        root = TreeNode(0, label='0')
        tree = Tree(root)
        node_list = [root]
        nr, leaf_count = 1, 1
        
        while leaf_count < N:
            node = random.choice(node_list)
            
            if not node.children: 
                # to be phylogenetic at least two children must be added
                new_child1 = TreeNode(nr, label=str(nr))
                new_child2 = TreeNode(nr+1, label=str(nr+1))
                node.add_child(new_child1)
                node.add_child(new_child2)
                node_list.extend(node.children)
                nr += 2
                leaf_count += 1
            elif node.children and not binary:
                # add only one child if there are already children
                new_child = TreeNode(nr, label=str(nr))
                node.add_child(new_child)
                node_list.append(new_child)
                nr += 1
                leaf_count += 1
                
        return tree


# -*- coding: utf-8 -*-

import time
from multiprocessing import Pool

from asymmetree.paraphylo.SpeciesTreeFromParalogs import TreeReconstructor
from asymmetree.file_io.ProteinOrthoParser import parse_po_graph


__author__ = 'David Schaller'


def _reconstruct(params):
    
    G, cotree_mode, triple_mode = params
    
    start_time = time.time()
    tr = TreeReconstructor(cotree_mode=cotree_mode)
    tr.add_ortho_graph(G)
    tr.build_species_tree(mode=triple_mode)
    newick = tr.newick_with_support()
    
    return newick, time.time() - start_time


def reconstruct_trees_and_write(infile, outfile, cotree_modes, triple_modes,
                                parallel_processing=True):
    
    G = parse_po_graph(infile)
    
    inputs = [(G, m1, m2) for m1 in cotree_modes for m2 in triple_modes]
    
    if parallel_processing:
        with Pool() as p:
            results = p.map(_reconstruct, inputs)
    else:
        results = list(map(_reconstruct, inputs))
    
    with open(outfile, 'w') as f:
        for i in range(len(inputs)):
            
            _, cotree_mode, triple_mode = inputs[i]
            newick, time_needed = results[i]
            f.write('# Cotree usage mode: {}, Max. Consistent Triple Set '\
                    'heuristic: {}, Time: {}\n'.format(cotree_mode,
                                                       triple_mode,
                                                       time_needed))
            f.write(newick + '\n')
            
            
def reconstruct_from_proteinortho(filename, triple_mode='BPMF'):
    
    G = parse_po_graph(filename)
    tr = TreeReconstructor()
    tr.add_ortho_graph(G)
    S = tr.build_species_tree(mode=triple_mode)
    support_newick = tr.newick_with_support()
    
    return S, support_newick


# -*- coding: utf-8 -*-

""" 
Parser for ProteinOrtho files.

Reads two different versions of ProteinOrtho result files (with/without
'-synteny' option) and builds a graph from the data (color --> 
species/filename).

References:
    - Marcus Lechner, Sven Findeiß, Lydia Steiner, Manja Marz, Peter F.
      Stadler, Sonja J. Prohaska. Proteinortho: Detection of (Co-)orthologs
      in large-scale analysis. BMC Bioinformatics 2011, 12:124
"""

import networkx as nx   


__author__ = 'David Schaller'


def parse_po_graph(filename):
    
    G = nx.Graph()
    
    with open(filename, 'r') as f:
        
        # skip the first two lines
        f.readline(); f.readline
        
        col_a, col_b = 'unknown', 'unknown'
        line = f.readline().strip()
        
        while line:
            
            line = line.strip()
            
            # new color pair begins
            if line and line.startswith('#') and not line.startswith('# Scores:'):
                colors = line.split()
                if len(colors) == 3:
                    col_a, col_b = colors[1], colors[2]
                
            elif line and not line.startswith('#'):
                edge = line.split()
                id1 = '{}_{}'.format(col_a, edge[0])
                id2 = '{}_{}'.format(col_b, edge[1])
                if id1 not in G:
                    G.add_node(id1, color=col_a)
                if id2 not in G:
                    G.add_node(id2, color=col_b)
                    
                # file type 1 with simscore (synteny was activated)
                if len(edge) == 8:
                    G.add_edge(id1, id2,
                               evalue_ab = float(edge[2]), 
                               bitscore_ab = float(edge[3]),
                               evalue_ba = float(edge[4]),
                               bitscore_ba = float(edge[5]),
                               same_strand = int(edge[6]),
                               simscore = float(edge[7]))
                
                # file type 2 without simscore
                elif len(edge) == 6:
                    G.add_edge(id1, id2,
                               evalue_ab = float(edge[2]), 
                               bitscore_ab = float(edge[3]),
                               evalue_ba = float(edge[4]),
                               bitscore_ba = float(edge[5]))
                else:
                    print('Check file format!', edge)
                    continue
                
            line = f.readline()
            
    return G


def parse_best_match_candidates(filename):
    
    candidates = {}
    
#    # extract the species from filename
#    species_match = re.search(r"([^/]+)\.bm_candidates$", filename)
#    if species_match:
#        species = species_match.group(1)
#    else:
#        raise ValueError("could not extract species from filename '{}'".format(filename))
    
    with open(filename, 'r') as f:
        
        line = f.readline()
        
        while line:
            
            line = line.strip()
            if not line:
                line = f.readline()
                continue
            
            data = line.split()
            
            if len(data) != 6:
                print('Check file format: {}, line {}'.format(filename, line))
                continue
                
            color      = data[0]
            query_id   = data[1]
            subject_id = data[2]
            e_value    = float(data[3])
            bitscore   = float(data[4])
            identity   = float(data[5]) / 100
            
            if subject_id not in candidates:
                candidates[subject_id] = {}
            if color not in candidates[subject_id]:
                candidates[subject_id][color] = []
            candidates[subject_id][color].append( (color, query_id, e_value, bitscore, identity) )
                
            line = f.readline()
            
    return candidates


def parse_best_matches(filename):
    
    G = nx.DiGraph()
    
    with open(filename, 'r') as f:
        
        line = f.readline()
        
        while line:
            
            line = line.strip().split()
            
            if len(line) == 4:
                
                col1, col2 = line[2], line[3]
                id1 = '{}_{}'.format(col1, line[0])
                id2 = '{}_{}'.format(col2, line[1])
                if id1 not in G:
                    G.add_node(id1, color=col1)
                if id2 not in G:
                    G.add_node(id2, color=col2)
                G.add_edge(id1, id2)
            
            line = f.readline()
            
    return G


# -*- coding: utf-8 -*-

import numpy as np
from scipy.special import zeta


__author__ = 'David Schaller'


class IndelModel:
    
    
    def __init__(self, insertion_rate, deletion_rate,
                 length_model='zipf',
                 max_length=False,
                 **kwargs):
        
        if insertion_rate < 0.0 or deletion_rate < 0.0:
            raise ValueError("insertion and deletion rates must be non-negative")
        else:
            self._ins_rate = insertion_rate     # insertion rate per site
            self._del_rate = deletion_rate      # deletion rate per site
        
        length_model = length_model.lower()
        if length_model in ('zipf', 'negative_binomial'):
            self._length_model = length_model
        else:
            raise ValueError("indel model '{}' is not available".format(length_model))
        
        self._max_length = max_length
        self._params = kwargs
        
        # check the parameter according to length model
        self._check_params()
        
        # compute(/set) mean for the indel length distribution
        self._del_mean = self._distribution_mean()
        
    
    def get_rates(self, seq_length):
        """Return the current insertion and deletion rate.
        
        Computes the current insertion and deletion rate according to the model
        and the length of the sequence."""
        
        return ( (seq_length + 1) * self._ins_rate,
                 (seq_length + self._del_mean - 1) * self._del_rate )
        
    
    def draw_length(self):
        """Draw the length for an indel."""
        
        while True:
            
            if self._length_model == 'zipf':
                d = np.random.zipf( self._params['a'] )
                    
            elif self._length_model == 'negative_binomial':
                d = 1 + np.random.negative_binomial(self._params['r'], 1 - self._params['q'])
            
            if self._max_length is False or d <= self._max_length:
                break
            
        return d
            
    
    def _check_params(self):
        
        if self._max_length is not False:
            
            if not isinstance(self._max_length, int) or self._max_length < 1:
                raise ValueError("maximal indel length must be an int >0")
        
        if self._length_model == 'zipf':
            
            if 'a' not in self._params:
                self._params = {'a': 1.821}     # (Chang and Benner 2004)
                
            elif not isinstance(self._params['a'], float) or self._params['a'] <= 1.0:
                raise ValueError("invalid value for parameter 'a': {}".format(self._params['a']))
                
        if self._length_model == 'negative_binomial':
            
            if 'r' not in self._params:
                self._params['r'] = 1
            if 'q' not in self._params:
                self._params['q'] = 0.5
                
            if not isinstance(self._params['r'], int) or self._params['r'] < 1:
                raise ValueError("parameter 'r' must be an int and >0")
            
            if (not isinstance(self._params['q'], float) or 
                self._params['q'] <= 0 or 
                self._params['q'] >= 1.0):
                raise ValueError("parameter 'q' must be a float >0 and <1")
            
            
    def _distribution_mean(self):
        
        if self._length_model == 'zipf':
            
            if self._params['a'] > 2.0:     # mean value is infinite for a <= 2
                return zeta(self._params['a'] - 1.0) / zeta(self._params['a'])
            else:
                return 0.0                  # maybe replace by some value > 0 ?
                
        if self._length_model == 'negative_binomial':
            return 1 + self._params['r'] * self._params['q'] / (1 - self._params['q'])
            


# -*- coding: utf-8 -*-

# best match computation
from asymmetree.best_matches.TrueBMG import orthology_from_tree, bmg_from_tree
from asymmetree.best_matches.LRTConstructor import lrt_from_observable_tree, LRTConstructor
from asymmetree.best_matches.Augmentation import augment_and_label

# best match estimation
from asymmetree.best_matches.ExtBestHits import ebh, ebh_qinfer, ebh_from_scenario
from asymmetree.best_matches.TreeReconstruction import bmg_by_tree_reconstruction, neighbor_joining, midpoint_rooting, nj_from_numpy_matrix
from asymmetree.best_matches.Quartets import Quartets, quartet_qinfer, quartet_from_scenario


# -*- coding: utf-8 -*-

import numpy as np
import networkx as nx


__author__ = 'David Schaller'


class AlignmentBuilder:
    
    
    def __init__(self, tree, sequence_dict, alphabet, include_inner=True):
        
        self.tree = tree
        self.sequence_dict = sequence_dict
        self.include_inner = include_inner
        
        if alphabet[-1] == '-':
            self.alphabet = alphabet
        else:
            self.alphabet = alphabet + '-'
        
        
    def build(self):
        
        self._get_preorder()
        self._sort_sites()
        self._alignment_matrix()
        
        return self._sequences()
        
    
    def _get_preorder(self):
        
        if self.include_inner:
            self.nodes = [v for v in self.tree.preorder()]
        else:
            self.nodes = [v for v in self.tree.preorder() if not v.children]
            
        self.node_index = {item: index for index, item in enumerate(self.nodes)}
        
    
    def _sort_sites(self):
        
        G = nx.DiGraph()
        
        for node in self.nodes:
            
            seq = self.sequence_dict[node]
            
            # handle case |seq| = 1 correctly
            if len(seq) == 1:
                G.add_node(seq[0].site_id)
            
            # add nodes and edges to auxiliary graph
            for first, second in seq.element_pairs():
                G.add_edge(first.site_id, second.site_id)
                
        self.sites = list(nx.topological_sort(G))
                
    
    def _alignment_matrix(self):
        
        self.alignment = np.zeros((len(self.nodes), len(self.sites)),
                             dtype=np.int8)
        positions = [self.sequence_dict[v]._first for v in self.nodes]
        
        for j in range(len(self.sites)):
            for i in range(len(positions)):
                
                if positions[i] and self.sites[j] == positions[i].site_id:
                    self.alignment[i, j] = positions[i]._value
                    positions[i] = positions[i]._next
                else:
                    self.alignment[i, j] = -1
    
    
    def _sequences(self):
        
        result = {}
        
        for i in range(len(self.nodes)):
            sequence = "".join(self.alphabet[x] for x in self.alignment[i, :])
            result[self.nodes[i]] = sequence
        
        return result


# -*- coding: utf-8 -*-

import numpy as np

from asymmetree.seqevolve.EvolvingSequence import State


__author__ = 'David Schaller'


class HetModel:
    
    def __init__(self, alpha, classes=5, sitewise=False, invariant=0.0):
        
        if not isinstance(alpha, float) or alpha <= 0.0:
            raise ValueError('heterogeneity parameter alpha must be a float > 0.0')
        self._alpha = alpha
        
        
        if not isinstance(classes, int) or classes <= 0:
            raise ValueError('number of classes must be an int >0')
        self.classes = classes
        
        self.sitewise = sitewise
        
        if not isinstance(invariant, float) or invariant < 0.0 or invariant > 1.0:
            raise ValueError('proportion of invariant sites must be in [0.0, 1.0]')
        self._invariant = invariant
        
        if not self.sitewise:
            self._initilize_classes()
        
    
    def _initilize_classes(self):
        
        if self.classes == 1:
            self._class_rates = [1.0]
            
        else:
            self._class_rates = np.random.gamma(self._alpha, scale=1/self._alpha,
                                                size=self.classes)
        
        
    def assign(self, sequence, exclude_inherited=True):
        
        n = len(sequence)
        if exclude_inherited:
            n -= sequence.count_status(State.INHERITED)
        
        rate_classes, rate_factors = self._draw(n)
        
        pos = 0
        for site in sequence:
            
            if not exclude_inherited or site.status != State.INHERITED:
                site.rate_class = rate_classes[pos] if rate_classes[pos] is not None else site.site_id
                site.rate_factor = rate_factors[pos]
                pos += 1
                    
                    
    def _draw(self, n):
        
        # mode 1: sitewise heterogeneity
        if self.sitewise:
            
            rate_classes = [None for _ in range(n)]
            rate_factors = np.random.gamma(self._alpha, scale=1/self._alpha, size=n)
        
        # mode 2: one or multiple  classes
        else:
            
            rate_classes = np.random.randint(self.classes, size=n).tolist()
            rate_factors = np.asarray([self._class_rates[c] for c in rate_classes])
            
        if self._invariant:
            
            drawn_variable = np.random.random(n) > self._invariant
            
            for i in range(n):
                if not drawn_variable[i]:
                    rate_classes[i] = 'invariant'
            
            rate_factors *= drawn_variable
        
        return rate_classes, rate_factors


# -*- coding: utf-8 -*-

"""
Species Tree Simulator.

Simulate dated species trees.
"""

import random, sys
import numpy as np

from asymmetree.tools.PhyloTree import (PhyloTree, PhyloTreeNode,
                                        delete_losses_and_contract,
                                        remove_planted_root)


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                         USER INTERFACE FUNCTION
# --------------------------------------------------------------------------

def simulate_species_tree(N, model='innovation',
                          non_binary_prob=0.0,
                          planted=True,
                          remove_extinct=False,
                          rescale_to_height=None,
                          **kwargs):
    """Simulates a species tree S with N leaves.
    
    Keyword parameters:
        model -- simulation model to be applied; default is 'innovation'
        non_binary_prob -- probability that an inner edge is contracted;
            results in non-binary tree; default is 0.0
        planted -- add a planted root that has the canonical root as its
            single neighbor; default is True
        remove_extinct -- remove all branches that lead to extinctions, only
            relevant for some models; default is False
        rescale_to_height -- determines the final distance from the root to the
            (surviving) leaves, default is None, i.e. model dependent
    """
    
    # parameter checking
    if not isinstance(N, int) or N < 0:
        raise ValueError('N must be an int >=0')
    elif N == 0:
        return PhyloTree(None)
    
    if not isinstance(model, str):
        raise ValueError("model must be of type 'str'")
    
    if non_binary_prob < 0.0 or non_binary_prob > 1.0:
        raise ValueError('contraction prob. must be in [0.0, 1.0]')
        
    if (rescale_to_height is not None and
        (not isinstance(rescale_to_height, (int, float)) or
         rescale_to_height < 0.0)):
        raise ValueError('height must be a number >=0')
    elif rescale_to_height is not None and N == 1 and not planted:
        raise ValueError('rescaling is not applicable to unplanted trees '\
                         'with only one leaf')
    
    # main simulation algorithm
    if model.lower() in ('innovation', 'innovations'):
        tree = _innovation_model(N, planted)
    elif model.lower() == 'yule':
        tree = _yule(N, kwargs.get('birth_rate'))
    elif model.upper() == 'BDP':
        tree = _BDP(N, **kwargs)
    elif model.upper() == 'EBDP':
        tree = _EBDP(N, **kwargs)
    else:
        raise ValueError("model '{}' is not available".format(model))
        
    # remove extinct branches for model that include losses
    if remove_extinct and model.upper() in ('BDP', 'EBDP'):
        delete_losses_and_contract(tree, inplace=True)
        
    # remove planted edge for models that are planted by construction
    if not planted and model.upper() in ('YULE', 'BDP', 'EBDP'):
        remove_planted_root(tree, inplace=True)
    
    # make tree non_binary by random contraction of edges
    if non_binary_prob > 0.0:
         edges = _select_edges_for_contraction(tree, non_binary_prob,
                                               exclude_planted_edge=True)
         tree.contract(edges)
        
    return tree


def simulate_species_tree_age(age, model='yule',
                              non_binary_prob=0.0,
                              **kwargs):
    """Simulates a (planted) species tree S of the specified age.
    
    Keyword parameters:
        model -- simulation model to be applied; default is 'yule'
        non_binary_prob -- probability that an inner edge is contracted;
                 results in non-binary tree; default is 0.0
    """
    
    # parameter checking
    if not isinstance(age, (float, int)) or age <= 0.0:
        raise ValueError('age must be a number >0')
    elif isinstance(age, int):
        age = float(age)
        
    if not isinstance(model, str):
        raise ValueError("model must be of type 'str'")
        
    if non_binary_prob < 0.0 or non_binary_prob > 1.0:
        raise ValueError("contraction prob. must be in [0.0, 1.0]")
    
    # main simulation algorithm
    if model.lower() == 'yule':
        tree = _yule_age(age, kwargs.get('birth_rate'))
    elif model.upper() == 'BDP':
        tree = _BDP_age(age, **kwargs)
    elif model.upper() == 'EBDP':
        tree = _EBDP_age(age, **kwargs)
    else:
        raise ValueError("model '{}' is not available".format(model))
        
    # make tree non_binary by random contraction of edges
    if non_binary_prob > 0.0:
         edges = _select_edges_for_contraction(tree, non_binary_prob,
                                               exclude_planted_edge=True)
         tree.contract(edges)
        
    return tree


# --------------------------------------------------------------------------
#                         AUXILIARY FUNCTIONS 
# --------------------------------------------------------------------------


def _rescale(tree, height, inplace=True):
    
    if not inplace:
        tree = tree.copy()
    
    old_height = tree.root.tstamp
    
    # not available for trees that only consist of a root
    if old_height <= 0.0:
        raise RuntimeError("cannot rescale tree of "\
                           "height '{}'".format(old_height))
        
    scaling_factor = height / old_height
    
    for v in tree.preorder():
        v.tstamp *= scaling_factor
        v.dist   *= scaling_factor
    
    return tree
            
            
def _select_edges_for_contraction(tree, p, exclude_planted_edge=True):
    
    edges = []
    
    for u, v in tree.inner_edges():
        
        if exclude_planted_edge and (u is tree.root) and len(u.children) == 1:
            continue
        
        if random.random() < p:
            edges.append((u,v))
    
    return edges


# --------------------------------------------------------------------------
#                     SPECIES TREE MODEL FUNCTIONS 
# --------------------------------------------------------------------------


def _innovation_model(N, planted, ultrametric=True):
    """Builds a species tree S with N leaves with the innovation model.
    
    Keyword arguments:
        ultrametric - if True make tree ultrametric and rescale it to
            height 1.0, else all edges have length 1.0; default is True
    """
    
    tree = PhyloTree(PhyloTreeNode(0, label='0'))
    tree.number_of_species = N
    node_counter = 1
    
    # planted tree (root is an implicit outgroup with outdegree = 1)
    if planted:
        root = PhyloTreeNode(1, label="1")
        tree.root.add_child(root)
        node_counter += 1
    else:
        root = tree.root
    
    features = [0]                  # set of available features
    species = {(0,): root}          # extant species
    
    while len(species) < N:
        
        loss_candidates = set()     # species for which loss of a feature
        for s in species.keys():    # can trigger a speciation
            for i in range(0, len(s)):
                    if s[:i] + s[i+1:] not in species:
                        loss_candidates.add(s)
        
        if not loss_candidates:     # INNOVATION EVENT
            s = random.choice(list(species))
            new_feature = len(features)
            
            new_s = s + (new_feature,)
            
            child1 = PhyloTreeNode(node_counter, label=str(node_counter))
            species[s].add_child(child1)
            child2 = PhyloTreeNode(node_counter+1, label=str(node_counter+1))
            species[s].add_child(child2)
            
            node_counter += 2
            
            species[s] = child1
            species[new_s] = child2
            features.append(new_feature)
            
        else:
            s = random.choice(list(loss_candidates))
            
            if len(s) > 1:
                feature_index = random.randint(0, len(s)-1)
            else:
                feature_index = 0
            
            new_s = s[:feature_index] + s[feature_index+1:]
            
            if new_s not in species:    # LOSS EVENT
                
                child1 = PhyloTreeNode(node_counter, label=str(node_counter))
                species[s].add_child(child1)
                child2 = PhyloTreeNode(node_counter+1, label=str(node_counter+1))
                species[s].add_child(child2)

                node_counter += 2
                
                species[s] = child1
                species[new_s] = child2
                
    if ultrametric:         
        _make_ultrametric(tree)
    
    return tree


def _make_ultrametric(tree):
    """Makes a given species tree S ultrametric.
    
    It is t(root) = 1 and t(x) = 0 for x in L(S)."""
    
    if not tree.root.children:
        tree.root.dist = 0.0
        tree.root.tstamp = 0.0
    
    for v in tree.preorder():
        if not v.parent:
            v.dist = 0.0
            v.tstamp = 1.0
        elif not v.children:
            v.dist = v.parent.tstamp
            v.tstamp = 0.0
        else:                               # random walk to a leaf
            pos = v                         # current position
            length = 0                      # path length |P|
            while pos.children:
                length += 1
                pos = pos.children[np.random.randint(len(pos.children))]
            v.dist = (v.parent.tstamp) * 2 * np.random.uniform() / (length+1)
            v.tstamp = v.parent.tstamp - v.dist


def _reverse_time_stamps(tree):
    
    max_depth = 0.0
    for v in tree.preorder():
        max_depth = max(max_depth, v.tstamp)
            
    for v in tree.preorder():
        v.tstamp = abs(v.tstamp - max_depth)


def _yule(N, birth_rate):
    
    if birth_rate is None:
        birth_rate = 1.0
    elif birth_rate <= 0.0:
        raise ValueError("birth rate must be >0")
    
    tree = PhyloTree(PhyloTreeNode(0, label='0', dist=0.0, tstamp=0.0))
    tree.number_of_species = N
    
    branches = [(1, tree.root)]
    forward_time = 0.0
    node_counter = 1
    
    while len(branches) < N:
        
        rate = len(branches) * birth_rate
        forward_time += np.random.exponential(1/rate)
        
        i = np.random.randint(len(branches))
        branch_id, parent = branches[i]
        spec_node = PhyloTreeNode(branch_id, label=str(branch_id),
                                  dist=forward_time-parent.tstamp,
                                  tstamp=forward_time)
        parent.add_child(spec_node)
        branches[i] = (node_counter, spec_node)
        branches.append((node_counter+1, spec_node))
        node_counter += 2
        
    # add length for pendant branches (cf. Hartmann et al. 2010)
    forward_time += np.random.exponential(1/rate)
    
    # finalize the branches
    for branch_id, parent in branches:
        parent.add_child( PhyloTreeNode(branch_id, label=str(branch_id),
                                        dist=forward_time-parent.tstamp,
                                        tstamp=forward_time) )
    
    _reverse_time_stamps(tree)
    
    return tree


def _yule_age(age, birth_rate):
    
    if birth_rate is None:
        birth_rate = 1.0
    elif birth_rate <= 0.0:
        raise ValueError("birth rate must be >0")
    
    tree = PhyloTree(PhyloTreeNode(0, label='0', dist=0.0, tstamp=0.0))
    
    branches = [(1, tree.root)]
    forward_time = 0.0
    node_counter = 1
    
    while forward_time < age:
        
        rate = len(branches) * birth_rate
        forward_time += np.random.exponential(1/rate)
        
        # do not branch if age is already reached
        if forward_time >= age:
            break
        
        i = np.random.randint(len(branches))
        branch_id, parent = branches[i]
        spec_node = PhyloTreeNode(branch_id, label=str(branch_id),
                                  dist=forward_time-parent.tstamp,
                                  tstamp=forward_time)
        parent.add_child(spec_node)
        branches[i] = (node_counter, spec_node)
        branches.append((node_counter+1, spec_node))
        node_counter += 2
    
    # finalize the branches
    for branch_id, parent in branches:
        parent.add_child( PhyloTreeNode(branch_id, label=str(branch_id),
                                        dist=age-parent.tstamp,
                                        tstamp=age) )
    
    # reverse such that t(root) = age and t(leaves) = 0.0
    _reverse_time_stamps(tree)
    
    return tree


def _BDP(N, **kwargs):
    
    # remove potentially supplied 'episodes' argument
    episodes = _EBDP_check_episodes(birth_rate = kwargs.get('birth_rate'),
                                    death_rate = kwargs.get('death_rate'))
    
    return _EBDP_backward(N, episodes)


def _EBDP(N, **kwargs):
    
    episodes = _EBDP_check_episodes(**kwargs)
    
    return _EBDP_backward(N, episodes)


def _EBDP_check_episodes(**kwargs):
    
    birth_rate = kwargs.get('birth_rate')
    death_rate = kwargs.get('death_rate')
    episodes = kwargs.get('episodes')
    
    # episodes parameter is prefered
    if episodes is not None:
        
        if len(episodes) == 0:
            raise ValueError("list of episodes must not be empty")
        
        for i in range(len(episodes)):
            
            if len(episodes[i]) != 4:
                raise ValueError("all episodes must contain 4 values: birth rate, "\
                                 "death rate, proportion of survivors, time stamp "\
                                 "(from recent time as 0)")
            
            birth_rate, death_rate, rho, t = episodes[i]
            if i == 0 and t != 0.0:
                raise ValueError("first episode must be at t=0.0")
            elif i > 0 and episodes[i-1][3] >= t:
                print(episodes[i-1][3], t)
                raise ValueError("episodes must be in correct temporal order")
            
            if birth_rate <= 0.0 or birth_rate < death_rate:
                raise ValueError("birth rate must be >0 and >=death rate "\
                                 "in all episodes")
                
            if rho <= 0.0 or rho > 1.0:
                raise ValueError("proportion of survivors must be in (0.0, 1.0]")
        
        return episodes
    
    elif birth_rate is not None:
        
        if birth_rate <= 0.0 or (death_rate and birth_rate < death_rate):
            raise ValueError("birth rate must be >0 and >=death rate")
            
        if death_rate and death_rate < 0.0:
            raise ValueError("death rate must be >=0")
        
        if death_rate is None:
            return [(birth_rate, 0.0, 1.0, 0.0)]
        else:
            return [(birth_rate, death_rate, 1.0, 0.0)]
        
    else:
        if death_rate:
            raise ValueError("birth rate (>0) must be specified if death rate "\
                             "is supplied")
            
        return [(1.0, 0.0, 1.0, 0.0)]


def _EBDP_backward(N, episodes, max_tries=500):
    """Episodic birth–death process (EBDP), backward algorithm by Stadler 2001."""
    
    birth_inv_sum = sum([1/episodes[i][0] for i in range(len(episodes))])
    
    for _ in range(max_tries):
        
        tree = None
        t = 0.0
        i = 0
        
        branches = [PhyloTreeNode(j, str(j), dist=0.0, tstamp=t)
                    for j in range(N)]
        id_counter = N
        
        while branches:
            birth_i, death_i, rho_i, t_i = episodes[i]
            
            losses_to_add = round(len(branches) / rho_i) - len(branches)
            for j in range(losses_to_add):
                branches.append( PhyloTreeNode(id_counter+j, '*', dist=0.0, tstamp=t) )
            id_counter += losses_to_add
            
            while branches:
                w = np.random.exponential( 1 / ((birth_i + death_i) * len(branches)) )
                
                if i+1 < len(episodes) and t + w > episodes[i+1][3]:
                    t = episodes[i+1][3]
                    i += 1
                    break
                
                else:
                    t += w
                    
                    if birth_i > np.random.uniform(low=0.0, high=birth_i+death_i):
                        # speciation event drawn
                        spec_node = PhyloTreeNode(id_counter, label=str(id_counter),
                                                  dist=0.0, tstamp=t)
                        id_counter += 1
                        if len(branches) > 1:
                            k, l = np.random.choice(len(branches), 2,
                                                    replace=False)
                            if k > l:
                                k, l = l, k
                            spec_node.add_child(branches[k])
                            spec_node.add_child(branches[l])
                            branches[k] = spec_node
                            branches.pop(l)
                        else:
                            spec_node.add_child(branches[0])
                            tree = PhyloTree(spec_node)
                            branches.clear()
                    else:
                        # extinction event drawn
                        branches.append( PhyloTreeNode(id_counter, '*',
                                                       dist=0.0, tstamp=t) )
                        id_counter += 1
        
        # return tree with the following probability
        if np.random.random() < (1 / birth_i) / birth_inv_sum:
            
            for v in tree.preorder():
                if v.parent:
                    v.dist = v.parent.tstamp - v.tstamp
            
            return tree
        
    print("Could not return a tree after {} simulations!".format(max_tries),
          file=sys.stderr)
   

def _BDP_age(age, **kwargs):
    
    # remove potentially supplied 'episodes' argument
    episodes = _EBDP_age_check_episodes(birth_rate = kwargs.get('birth_rate'),
                                        death_rate = kwargs.get('death_rate'))
    
    return _EBDP_age_forward(age, episodes)


def _EBDP_age(age, **kwargs):
    
    episodes = _EBDP_age_check_episodes(**kwargs)
    
    return _EBDP_age_forward(age, episodes)


def _EBDP_age_check_episodes(**kwargs):
    
    birth_rate = kwargs.get('birth_rate')
    death_rate = kwargs.get('death_rate')
    episodes = kwargs.get('episodes')
    
    # episodes parameter is prefered
    if episodes is not None:
        
        if len(episodes) == 0:
            raise ValueError("list of episodes must not be empty")
        
        for i in range(len(episodes)):
            
            if len(episodes[i]) != 4:
                raise ValueError("all episodes must contain 4 values: birth rate, "\
                                 "death rate, proportion of survivors, time stamp "\
                                 "(from recent time as 0)")
            
            birth_rate, death_rate, rho, t = episodes[i]
            if i == 0 and t != 0.0:
                raise ValueError("first episode must be at t=0.0")
            elif i > 0 and episodes[i-1][3] >= t:
                print(episodes[i-1][3], t)
                raise ValueError("episodes must be in correct temporal order")
            
            if birth_rate < 0.0 or death_rate < 0.0:
                raise ValueError("birth and death rate must be >=0 "\
                                 "in all episodes")
                
            if rho <= 0.0 or rho > 1.0:
                raise ValueError("proportion of survivors must be in (0.0, 1.0]")
        
        return episodes
    
    elif birth_rate is not None:
        
        if birth_rate < 0.0 or (death_rate is not None and death_rate < 0.0):
            raise ValueError("birth and death rate must be >=0")
        
        if death_rate is None:
            # default death rate = 0.0
            return [(birth_rate, 0.0, 1.0, 0.0)]
        else:
            return [(birth_rate, death_rate, 1.0, 0.0)]
        
    else:
        if death_rate:
            raise ValueError("birth rate (>=0) must be specified if death rate "\
                             "is supplied")
        
        # default birth rate = 1.0 and death rate = 0.0
        return [(1.0, 0.0, 1.0, 0.0)]

   
def _EBDP_mass_extinction(branches, surviving_rate, t):
    
    no_of_losses = round((1-surviving_rate) * len(branches))
    chosen_losses = sorted(np.random.choice(len(branches), replace=False,
                                            size=no_of_losses),
                           reverse=True)
    for j in chosen_losses:
        branch_id, parent = branches[j]
        loss_node = PhyloTreeNode(branch_id, '*',
                                  dist=t-parent.tstamp,
                                  tstamp=t)
        parent.add_child(loss_node)
        branches.pop(j)
    

def _EBDP_age_forward(age, episodes):
    """Episodic birth–death process (EBDP), forward algorithm with max. age."""
    
    tree = PhyloTree(PhyloTreeNode(0, label='0', dist=0.0, tstamp=0.0))
    
    branches = [(1, tree.root)]
    forward_time = 0.0
    node_counter = 1
    i = 0               # current episode
    
    # may lead to extinction of the single branch at time t=0
    _EBDP_mass_extinction(branches, episodes[i][2], episodes[i][3])
    
    while forward_time < age:
        birth_rate, death_rate, *_ = episodes[i]
        
        rate = len(branches) * (birth_rate + death_rate)
        waiting_time = np.random.exponential(1/rate) if rate > 0.0 else float('inf')
        
        if i+1 < len(episodes) and forward_time + waiting_time >= episodes[i+1][3]:
            _EBDP_mass_extinction(branches, episodes[i+1][2], episodes[i+1][3])
            forward_time = episodes[i+1][3]
            i += 1
        
        elif forward_time + waiting_time >= age:
            break
        
        else:
            forward_time += waiting_time
            
            j = np.random.randint(len(branches))
            branch_id, parent = branches[j]
            
            if birth_rate > np.random.uniform(low=0.0, high=birth_rate+death_rate):
                # speciation event drawn
                spec_node = PhyloTreeNode(branch_id, label=str(branch_id),
                                          dist=forward_time-parent.tstamp,
                                          tstamp=forward_time)
                parent.add_child(spec_node)
                branches[j] = (node_counter, spec_node)
                branches.append((node_counter+1, spec_node))
                node_counter += 2
            else:
                # extinction event drawn
                loss_node = PhyloTreeNode(branch_id, '*',
                                          dist=forward_time-parent.tstamp,
                                          tstamp=forward_time)
                parent.add_child(loss_node)
                branches.pop(j)
                
    # finalize the (surviving) branches
    for branch_id, parent in branches:
        parent.add_child( PhyloTreeNode(branch_id, label=str(branch_id),
                                        dist=age-parent.tstamp,
                                        tstamp=age) )
    
    # reverse such that t(root) = age and t(surviving leaves) = 0.0
    for v in tree.preorder():
        v.tstamp = age - v.tstamp
    
    return tree


import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='asymmetree',
    version='0.1.0',
    author='David Schaller',
    author_email='david@0x002A.com',
    description='Simulation of phylogenetic trees, sequences and genomes.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/david-schaller/AsymmeTree',
    packages=setuptools.find_packages(exclude=['examples', 'examples.*',
                                               'resources', 'resources.*',
                                               'tests', 'tests.*',
                                               'validation', 'validation.*']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'numpy>=1.16.4',
        'scipy>=1.3.0',
        'matplotlib>=3.0',
        'networkx>=2.2',
   ],
)



# -*- coding: utf-8 -*-

"""
Tree data structure for phylogentic trees.
"""

import collections, itertools, random, re, pickle, json, os
import numpy as np
import networkx as nx

from asymmetree.tools.Tree import Tree, TreeNode


__author__ = 'David Schaller'


class PhyloTreeNode(TreeNode):
    """Class 'PhyloTreeNode'.
    
    Components for class 'PhyloTree'.
    """
    
    __slots__ = ('color', 'dist', 'tstamp', 'transferred')
    
    
    def __init__(self, ID, label='', color=None,
                 dist = 1.0, tstamp=None, transferred=0):
        
        super().__init__(ID, label=label)
        self.color = color                      # color / reconciliation
        self.dist = dist                        # distance from parent
        self.tstamp = tstamp                    # time stamp
        self.transferred = transferred          # 1 if edge parent-self was an 
                                                # HGT edge, and 0 otherwise

    
    def __repr__(self):
        
        return '<PhyloTN: {}>'.format(self.ID)
    
    
    def __str__(self):
        
        if isinstance(self.color, (tuple, list)):
            return '{}<{}-{}>:{}'.format(self.label, *self.color, self.dist)
        elif self.color:
            return '{}<{}>:{}'.format(self.label, self.color, self.dist)
        else:
            return '{}:{}'.format(self.label, self.dist)
        
    
    def is_loss(self):
        
        return self.label == '*'


class PhyloTree(Tree):
    
    # corresponding node type
    node_type = PhyloTreeNode
    
    
    def __init__(self, root):
        
        if root is not None and not isinstance(root, PhyloTreeNode):
            raise TypeError("root must be of type 'PhyloTreeNode'")
        super().__init__(root)
    
    
    def sorted_nodes(self):
        """Return list of sorted nodes by timestamp."""
        
        return sorted(self.preorder(),
                      key=lambda node: node.tstamp,
                      reverse=True)
    
    
    def sorted_edges(self):
        """Return list of sorted edges (u,v) by timestamp of u."""
    
        return sorted(self.edges(),
                      key=lambda e: (e[0].tstamp, e[1].tstamp),
                      reverse=True)
    
    
    def delete_and_reconnect(self, node,
                             add_distances=True,
                             keep_transferred=True):
        """Delete a node from the tree and reconnect its parent and children."""
        
        distance, transferred = node.dist, node.transferred
        parent = super().delete_and_reconnect(node)
        
        if parent:
            for child in parent.children:
                if add_distances:
                    child.dist += distance
                if keep_transferred and transferred:
                    child.transferred = 1
        
        return parent
        
    
    def contract(self, edges):
        
        contracted = set()
        
        for u, v in edges:
            
            # avoid trying to contract the same edge multiple times
            if v not in contracted:
                self.delete_and_reconnect(v)
            
            contracted.add(v)
            
            
    def remove_planted_root(self):
        """Removes the planted root if existent.
        
        The the tree has a planted root, its single child becomes the new root.
        """
        
        if len(self.root.children) == 1:
        
            new_root = self.root.children[0]
            new_root.detach()
            self.root = new_root
            new_root.dist = 0.0
            
    
    def supply_leaves(self, exclude_losses=True):
        """Add the leaves to all nodes under their respective subtrees."""
        
        def _supply_leaves(node, exclude_losses):
        
            node.leaves = []
            
            if not node.children:
                if exclude_losses and not node.is_loss():
                    node.leaves.append(node)
                elif not exclude_losses:
                    node.leaves.append(node)
            else:
                for child in node.children:
                    node.leaves.extend(_supply_leaves(child, exclude_losses))
                    
            return node.leaves
        
        
        if self.root:
            return _supply_leaves(self.root, exclude_losses)
        else:
            return []
    
    
    def color_sorted_leaves(self):
        
        self.supply_leaves()
        color_dict = {}
        
        for leaf in self.root.leaves:
            if leaf.color not in color_dict:
                color_dict[leaf.color] = []
            color_dict[leaf.color].append(leaf)
            
        leaves = []
        for color, leaf_list in color_dict.items():
            for leaf in leaf_list:
                leaves.append(leaf)
        
        return leaves
    
    
    def distance_matrix(self, leaf_order=None):
        """Computes the distance matrix from the phylogenetic tree.
        
        Additionally a list of leaves corresponding to the indices in the matrix
        is returned.
        
        Keyword arguments:
            leaf_order -- list of leaves defining the indices for the matrix;
                default is None, in which case leaves are indexed in sibling order.
        """
        
        distance_dict, _ = self.distances_from_root()
        self.supply_leaves()
        
        if leaf_order:
            if set(leaf_order) != set(self.root.leaves):
                raise ValueError('ordered leaf list does not match with the '\
                                 'leaves in the tree')
            leaves = leaf_order
        else:
            # leaves in sibling order
            leaves = self.root.leaves      
        
        leaf_index = {l: i for i, l in enumerate(leaves)}
        
        D = np.zeros((len(leaves),len(leaves)), dtype=np.float)
        
        for v in self.preorder():
            if v.children:
                for c1, c2 in itertools.combinations(v.children, 2):
                    for x in c1.leaves:
                        x_index = leaf_index[x]
                        x_dist = distance_dict[x] - distance_dict[v]
                        for y in c2.leaves:
                            y_index = leaf_index[y]
                            y_dist = distance_dict[y] - distance_dict[v]
                            D[x_index, y_index] = x_dist + y_dist
                            D[y_index, x_index] = x_dist + y_dist
        
        return leaves, D
    
    
    def distances_from_root(self):
        
        leaf_distances = []
        distance_dict = {}
        
        for v in self.preorder():
            if not v.parent:
                distance_dict[v] = 0.0
            else:
                depth = distance_dict[v.parent] + v.dist
                distance_dict[v] = depth
            if not v.children:
                leaf_distances.append( (str(v), distance_dict[v]) )
                
        return distance_dict, leaf_distances
    
    
    def topology_only(self, inplace=True):
        """Reset distances, time stamps, transfer status and inner labels."""
        
        if not inplace:
            T = self.copy()
        else:
            T = self
        
        for v in T.preorder():
            if v.children:
                v.label = ""
                v.color = None
            v.dist = 1.0
            v.tstamp = None
            v.transferred = 0
        
        return T
    
    
    def count_node_types(self):
        
        counts = {'S': 0, 'D': 0, 'L': 0, 'H': 0, 'extant': 0}
        
        for v in self.preorder():
            
            if not v.children:
                if v.is_loss():
                    counts['L'] += 1
                else:
                    counts['extant'] += 1
            
            elif v.label == 'S':
                counts['S'] += 1
            elif v.label == 'D':
                counts['D'] += 1
            elif v.label == 'H':
                counts['H'] += 1
        
        return counts        
    

# --------------------------------------------------------------------------
#                          TREE  <--->  NEWICK
# --------------------------------------------------------------------------
        
    def to_newick(self, label=True, color=True, distance=True,
                        label_inner=True, color_inner=False):
        """Recursive PhyloTree --> Newick (str) function."""
        
        def _to_newick(node):
            
            if not node.children:
                token = ''
                if label:
                    token += str(node.label)
                if color and node.color:
                    token += '<{}-{}>'.format(*node.color) if isinstance(node.color, (tuple, list)) else '<{}>'.format(node.color)
                if distance:
                    token += ":{}".format(node.dist)
                return token
            else:
                s = ''
                for child in node.children:
                    s += _to_newick(child) + ','
                token = ''
                if label and label_inner:
                    token += str(node.label)
                if color_inner and node.color:
                    token += '<{}-{}>'.format(*node.color) if isinstance(node.color, (tuple, list)) else '<{}>'.format(node.color)
                if distance:
                    token += ':{}'.format(node.dist)
                return '({}){}'.format(s[:-1], token)
        
        
        if self.root:
            return _to_newick(self.root) + ';'
        else:
            return ';'
    
    
    @staticmethod
    def parse_newick(newick):
        """Parses trees in Newick format into object of type 'PhyloTree'."""
        
        # label<color>:distance
        label_col_dist_regex = re.compile(r"'?([a-zA-Z0-9_]*)'?<(.*)>:(-?[0-9]*\.?[0-9]*[Ee]?-?[0-9]+)")
        # label<color>
        label_col_regex = re.compile(r"'?([a-zA-Z0-9_]*)'?<(.*)>")
        # label:distance
        label_dist_regex = re.compile(r"'?([a-zA-Z0-9_]*)'?:(-?[0-9]*\.?[0-9]*[Ee]?-?[0-9]+)")
        
        id_counter = 0
        
        def parse_subtree(subroot, subtree_string):
            """Recursive function to parse the subtrees."""
            nonlocal id_counter
            children = split_children(subtree_string)
            for child in children:
                node = PhyloTreeNode(id_counter)
                subroot.add_child(node)
                id_counter += 1
                end = -1
                if child[0] == '(':                                 # the child has subtrees
                    end = child.rfind(')')
                    if end == -1:
                        raise ValueError('invalid Newick string')
                    parse_subtree(node, child[1:end])               # recursive call 'parse_subtree'
                child = child[end+1:].strip()
                label_col_dist = label_col_dist_regex.match(child)
                if label_col_dist:                                  # CASE 1: label<color>:distance
                    node.label = label_col_dist.group(1)
                    node.color = label_col_dist.group(2)
                    node.dist = float(label_col_dist.group(3))
                else:
                    label_col = label_col_regex.match(child)
                    label_dist = label_dist_regex.match(child)
                    if label_col:                                   # CASE 2: label<color>
                        node.label = label_col.group(1)
                        node.color = label_col.group(2)
                    elif label_dist:                                # CASE 3: label:distance
                        node.label = label_dist.group(1)
                        node.dist = float(label_dist.group(2))
                    else:                                           # CASE 4: label
                        node.label = child
                if node.color and node.color.find('-') != -1:
                    split_color = node.color.split('-')
                    node.color = (split_color[0], split_color[1])
                        
        def split_children(child_string):
            """Splits a given string by all ',' that are not enclosed by parentheses."""
            
            stack = 0
            children = []
            current = ""
            for c in child_string:
                if (stack == 0) and c == ',':
                    children.append(current)
                    current = ""
                elif c == '(':
                    stack += 1
                    current += c
                elif c == ')':
                    if stack <= 0:
                        raise ValueError('invalid Newick string')
                    stack -= 1
                    current += c
                else:
                    current += c
            children.append(current.strip())
            return children
        
        if not isinstance(newick, str):
            raise ValueError("Newick parser needs a 'str' as input")
        end = newick.find(";")
        if end != -1:
            newick = newick[:end]
        temp_root = PhyloTreeNode(-1)
        parse_subtree(temp_root, newick)
        if temp_root.children:
            root = temp_root.children[0]
            root.dist = 0.0                 # set distance of the root to 0
            root.detach()                   # remove the parent temp_root
                                            # (important for non-recursive to_newick2)
            return PhyloTree(root)
        else:
            raise ValueError('invalid Newick string')
    
    
    def reconstruct_IDs(self):
        """Reconstruct the (leaf) IDs."""
        
        self.number_of_species = 0
        IDs = set()
        
        for v in self.preorder():
            if not v.children:
                self.number_of_species += 1
            if v.label.isdigit():
                v.ID = int(v.label)
                IDs.add(v.ID)
        
        # assign new IDs to internal nodes
        current_ID = 0
        for v in self.preorder():
            if str(v.ID) != v.label:
                while current_ID in IDs:
                    current_ID += 1
                v.ID = current_ID
                IDs.add(current_ID)
                
                
    def reconstruct_timestamps(self):
        """Reconstruct the timestamps."""
        
        self.root.tstamp = 1.0
        for v in self.preorder():
            if v.parent:
                v.tstamp = v.parent.tstamp - v.dist
    
# --------------------------------------------------------------------------
#                         TREE  <--->  NETWORKX
# --------------------------------------------------------------------------
            
    def to_nx(self):
        
        self._assert_integrity()
        G = nx.DiGraph()
        
        if not self.root:
            return G, None
        
        for v in self.preorder():
            G.add_node(v.ID, label=v.label,
                       color=v.color,
                       tstamp=v.tstamp,
                       dist=v.dist,
                       transferred=v.transferred)
        
        for u, v, sibling_nr in self.edges_sibling_order():
            if u is v or u.ID == v.ID:
                raise RuntimeError('loop at {} and {}'.format(u, v))
            G.add_edge(u.ID, v.ID)
            G.nodes[v.ID]['sibling_nr'] = sibling_nr
            
        return G, self.root.ID
    
    
    @staticmethod
    def parse_nx(G, root):
        
        number_of_leaves = 0
        
        if root is None:
            return PhyloTree(None)
    
        def build_tree(ID, parent=None):
            
            nonlocal number_of_leaves
            
            treenode = PhyloTreeNode(ID, label=G.nodes[ID]['label'])
            
            if parent:
                parent.add_child(treenode)
                
                # edge attributes are deprecated,
                # for compatibilty with older versions of AsymmeTree:
                if 'dist' in G.edges[parent.ID, ID]:
                    treenode.dist = G.edges[parent.ID, ID]['dist']
                if 'transferred' in G.edges[parent.ID, ID]:
                    treenode.transferred = G.edges[parent.ID, ID]['transferred']
            
            if 'label' in G.nodes[ID]:
                treenode.label = G.nodes[ID]['label']
            if 'color' in G.nodes[ID]:
                if isinstance(G.nodes[ID]['color'], list):
                    treenode.color = tuple(G.nodes[ID]['color'])
                else:
                    treenode.color = G.nodes[ID]['color']
            if 'tstamp' in G.nodes[ID]:
                treenode.tstamp = G.nodes[ID]['tstamp']
            if 'dist' in G.nodes[ID]:
                treenode.dist = G.nodes[ID]['dist']
            if 'transferred' in G.nodes[ID]:
                treenode.transferred = G.nodes[ID]['transferred']
            
            try:
                children_ids = sorted(G.neighbors(ID),
                                  key=lambda item: G.nodes[item]['sibling_nr'])
            except KeyError:
                # for compatibilty with older versions of AsymmeTree:
                children_ids = [item for item in G.neighbors(ID)]
            
            for child_id in children_ids:
                build_tree(child_id, parent=treenode)
            if G.out_degree(ID) == 0:
                number_of_leaves += 1
            
            return treenode
        
        tree = PhyloTree(build_tree(root))
        tree.number_of_species = number_of_leaves
        
        return tree

# --------------------------------------------------------------------------
#                           SERIALIZATION
# --------------------------------------------------------------------------
    
    @staticmethod
    def _infer_serialization_mode(filename):
        
        _, file_ext = os.path.splitext(filename)
        
        if file_ext in ('.json', '.JSON'):
            return 'json'
        elif file_ext in ('.pickle', '.PICKLE'):
            return 'pickle'
        else:
            raise ValueError('serialization format is not supplied and could '\
                             'not be inferred from file extension')
            
            
    def serialize(self, filename, mode=None):
        """Serialize the tree using pickle."""
        
        if not mode:
            mode = PhyloTree._infer_serialization_mode(filename)
        
        tree_nx, root_id = self.to_nx()
        
        if mode == 'json':
            data = nx.readwrite.json_graph.tree_data(tree_nx, root=root_id)
            
            with open(filename, 'w') as f:
                f.write( json.dumps(data) )
                
        elif mode == 'pickle':
            pickle.dump( (tree_nx, root_id), open(filename, "wb") )
            
        else:
            raise ValueError("serialization mode '{}' not supported".format(mode))
    
    
    @staticmethod
    def load(filename, mode=None):
        """Load a phylogenetic tree from a file.
        
        Using either the Python module json or pickle."""
        
        if not mode:
            mode = PhyloTree._infer_serialization_mode(filename)
        
        if mode == 'json':
            with open(filename, 'r') as f:
                data = json.loads( f.read() )
                
            tree_nx = nx.readwrite.json_graph.tree_graph(data)
            
            root_id = None
            for v in tree_nx:
                if tree_nx.in_degree(v) == 0:
                    root_id = v
                    break
                
            if root_id is None:
                raise RuntimeError('could not identify root')
                
        elif mode == 'pickle':
            tree_nx, root_id = pickle.load( open(filename, "rb") )
            
        else:
            raise ValueError("serialization mode '{}' not supported".format(mode))
        
        tree = PhyloTree.parse_nx(tree_nx, root_id)
        
        return tree
    
# --------------------------------------------------------------------------
        
    
    def copy(self):
        
        if not self.root:
            return PhyloTree(None)
        
        orig_to_new = {}
        
        for orig in self.preorder():
            new = PhyloTreeNode(orig.ID, label=orig.label, 
                                color=orig.color, dist=orig.dist,
                                tstamp=orig.tstamp,
                                transferred=orig.transferred)
            orig_to_new[orig] = new
            if orig.parent:
                orig_to_new[orig.parent].add_child(new)
        
        return PhyloTree(orig_to_new[self.root])
    
    
    @staticmethod
    def random_colored_tree(N, colors, binary=False):
        """Creates a random colored tree.
        
        The number of leaves and the color labels are specified in the
        parameters 'N' and 'colors', respectively. Each non-leaf node in the 
        resulting tree will have at least children (property of phylogenetic
        trees).
        
        Keyword arguments:
            binary - forces the tree to be binary; default is False
        """
        
        if not isinstance(N, int):
            raise TypeError("N must be of type 'int'")
            
        if isinstance(colors, int):
            colors = [i+1 for i in range(colors)]
        elif not isinstance(colors, collections.abc.Iterable):
            raise TypeError("'colors' must be of type 'int' or iterable")
        
        root = PhyloTreeNode(0, label='0')
        tree = PhyloTree(root)
        node_list = [root]
        nr, leaf_count = 1, 1
        
        while leaf_count < N:
            node = random.choice(node_list)
            
            if not node.children: 
                # to be phylogenetic at least two children must be added
                new_child1 = PhyloTreeNode(nr, label=str(nr))
                new_child2 = PhyloTreeNode(nr+1, label=str(nr+1))
                node.add_child(new_child1)
                node.add_child(new_child2)
                node_list.extend(node.children)
                nr += 2
                leaf_count += 1
            elif node.children and not binary:
                # add only one child if there are already children
                new_child = PhyloTreeNode(nr, label=str(nr))
                node.add_child(new_child)
                node_list.append(new_child)
                nr += 1
                leaf_count += 1
                
        for node in node_list:                          # assign colors
            if not node.children:                       # to leaves randomly
                node.color = random.choice(colors)
                
        return tree
   
     
# --------------------------------------------------------------------------
#                         TREE MANIPULATION
# -------------------------------------------------------------------------- 

def delete_losses_and_contract(tree, inplace=False):
    
    if not inplace:
        tree = tree.copy()
    
    loss_nodes = []
    for node in tree.postorder():
        if not node.children and node.is_loss():
            loss_nodes.append(node)
    
    # traverse from loss node to root delete if degree <= 1
    for loss_node in loss_nodes:
        current = tree.delete_and_reconnect(loss_node,
                                            add_distances=True,
                                            keep_transferred=True)
        
        while len(current.children) < 2 and current.parent:
            current = tree.delete_and_reconnect(current,
                                                add_distances=True,
                                                keep_transferred=True)
    
    return tree


def remove_planted_root(tree, inplace=True):
    
    if not inplace:
        tree = tree.copy()
        
    # delete the root if the tree is planted
    tree.remove_planted_root()
        
    if not tree.root.children and not tree.root.label:
        # no surviving genes --> return empty tree
        tree.root = None
    
    return tree


# -*- coding: utf-8 -*-

"""
Gene Tree Simulator.

Simulate dated gene trees.
"""

import random
from collections import deque

import numpy as np

from asymmetree.tools.PhyloTree import (PhyloTree, PhyloTreeNode,
                                        delete_losses_and_contract,
                                        remove_planted_root)


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                    CONSTRUCTION OF THE GENE TREE
# --------------------------------------------------------------------------
    

def simulate_dated_gene_tree(S, **kwargs):
    
    gene_tree_simulator = GeneTreeSimulator(S)
    return gene_tree_simulator.simulate(**kwargs)


class _Branch:
    
    __slots__ = ('ID', 'array_id', 'rate',
                 'parent', 'S_u', 'S_v', 'transferred')
    
    def __init__(self, ID, array_id, rate,
                 parent, S_u, S_v, transferred):
        
        self.ID = ID                                    # unique branch id
        self.array_id = array_id                        # index in rate array
        self.rate = rate                                # total event rate
        
        self.parent = parent                            # parent node
        self.S_u = S_u                                  # edge (S_u, S_v) into which
        self.S_v = S_v                                  # the branch is embedded
        self.transferred = transferred                  # 1 if HGT edge, 0 otherwise
        

class GeneTreeSimulator:
    
    def __init__(self, S):
        
        self.S = S                                      # species tree
        self.sorted_speciations = S.sorted_nodes()      # list of speciations sorted by time stamp
        self.sorted_edges = S.sorted_edges()            # edges of the species tree
                                                        # sort (u,v) by tstamp of u
    
    
    def simulate(self, DLH_rates=(0.0, 0.0, 0.0),
                 dupl_polytomy=0.0,
                 prohibit_extinction='per_species',
                 **kwargs):
        """Simulate a gene tree along the specified species tree.
        
        Keyword arguments:
            DLH_rates -- rates for duplications, losses and HGTs, default is
                (0.0, 0.0, 0.0)
            dupl_polytomy -- allows non-binary duplication events by specifying
                the lambda parameter for a poisson distribution (copy number =
                drawn number + 2); default is 0.0
            prohibit_extinction -- avoid the extinction of all members in any
                species ('per_species'), of the complete gene family
                ('per_family'), or no constraints (False); default is 
                'per_species'.
        """
        
        self.DLH_rates = DLH_rates
        self.rate_sum = sum(DLH_rates)
        self.d = DLH_rates[0]
        self.l = DLH_rates[1]
        self.h = DLH_rates[2]
        
        self._prohibit_extinction = prohibit_extinction
        
        self._dupl_polytomy = dupl_polytomy
        
        self._reset()
        
        return self._gillespie_simulation()
    
    
    def _reset(self):
        
        self.spec_queue = deque(self.sorted_speciations)    # queue for speciation events
        self.id_counter = 0
        
        self.total_surviving = 0                            # counter for surviving genes
        self.total_rate = 0                                 # total event rate (all branches)
        self.branches = []
        
        self.ES_to_b = {e: [] for e in self.sorted_edges}   # maps E(S) --> existing branches

    
    def _get_tstamp(self, t):
        
        if self._prohibit_extinction == 'per_family' and self.total_surviving == 1:
            rate = self.d + self.l
        else:
            rate = self.total_rate
        
        
        if rate <= 0.0:
            return -1
        else:
            return t - np.random.exponential(1/rate)
    
    
    def _get_branch_and_type(self):
        
        if self._prohibit_extinction == 'per_family' and self.total_surviving == 1:
            
            branch = None
            for b in self.branches:
                if b.rate > 0.0:
                    branch = b
                    break
            
            if np.random.uniform(high=self.d+self.h) <= self.d:
                event_type = 'D'
            else:
                event_type = 'H'
        
        else:
            r = np.random.uniform(high=self.total_rate)
            current_sum = 0.0
            
            for i in range(len(self.branches)):
                if r <= current_sum + self.branches[i].rate:
                    break
                current_sum += self.branches[i].rate
            
            branch = self.branches[i]
            
            loss_factor = 1
            if (self._prohibit_extinction == 'per_species' and
                len(self.ES_to_b[(branch.S_u,branch.S_v)]) <= 1):
                loss_factor = 0
            
            if r <= current_sum + self.d:
                event_type = 'D'
            elif r <= current_sum + self.d + loss_factor * self.l:
                event_type = 'L'
            else:
                event_type = 'H'
        
#        assert np.isclose(self.total_rate, sum([b.rate for b in self.branches])), "Sum of rates and total rate are not equal."
        
        return branch, event_type
    
    
    def _get_copy_number(self):
        
        if self._dupl_polytomy <= 0.0:
            return 2
        else:
            return 2 + np.random.poisson(lam=self._dupl_polytomy)


    def _coexisting_edges(self, tstamp, exclude_edge=None):
        """Return list of edges for the given timestamp."""
        
        valid_edges = []
        
        for edge in self.sorted_edges:
            
            if edge[0].tstamp <= tstamp:
                break
            elif edge[1].tstamp < tstamp and edge != exclude_edge:
                valid_edges.append(edge)
                
        return valid_edges


    def _gillespie_simulation(self):
        
        T = self._initiatialize_tree()
        t = T.root.tstamp                   # start time = 1.0
        
        while self.spec_queue:
            
            event_tstamp = self._get_tstamp(t)
            next_spec_tstamp = self.spec_queue[0].tstamp
            
            # speciation
            if event_tstamp <= next_spec_tstamp:
                self._speciation()      
                t = next_spec_tstamp
                
            else:
                branch, event_type = self._get_branch_and_type()
                
                # duplication
                if event_type == 'D':
                    self._duplication(event_tstamp, branch)
                
                # loss
                elif event_type == 'L':
                    self._loss(event_tstamp, branch)
                    
                # HGT    
                elif event_type == 'H':
                    self._hgt(event_tstamp, branch)
                
                t = event_tstamp

        return T
    
    
    def _initiatialize_tree(self):
        
        if len(self.S.root.children) > 1:
            # root is a speciation event
            root = PhyloTreeNode(self.id_counter, label='S', color=self.S.root.ID, 
                                 dist=0.0, tstamp=self.S.root.tstamp)
        else:                    
            # planted species tree
            root = PhyloTreeNode(self.id_counter, color=self.S.root.ID,
                                 dist=0.0, tstamp=self.S.root.tstamp)
            
        T = PhyloTree(root)
        self.id_counter += 1
        self.spec_queue.popleft()
        
        if self._prohibit_extinction == 'per_species':
            rate = self.d + self.h
        else:
            rate = self.rate_sum
                
        for S_v in self.S.root.children:
            
            array_id = len(self.branches)
            new_branch = _Branch(self.id_counter, array_id, rate,
                                 T.root, self.S.root, S_v, 0)
            self.ES_to_b[(self.S.root, S_v)].append(new_branch)
            self.branches.append(new_branch)
            self.id_counter += 1
            
        self.total_rate += len(self.S.root.children) * rate   
        self.total_surviving = len(self.S.root.children)
            
        return T
            
    
    def _speciation(self):
        
        S_v = self.spec_queue.popleft()
        S_u = S_v.parent
        
        for branch in self.ES_to_b[(S_u, S_v)]:
            
            spec_node = PhyloTreeNode(branch.ID, label='S',
                                      color=S_v.ID, tstamp=S_v.tstamp,
                                      dist=abs(S_v.tstamp-branch.parent.tstamp),
                                      transferred=branch.transferred)
            branch.parent.add_child(spec_node)
            
            if not S_v.children:
                spec_node.label = str(spec_node.ID)
                
            for S_w in S_v.children:
                
                if S_w is S_v.children[0]:
                    array_id = branch.array_id
                    new_branch = _Branch(self.id_counter, array_id, branch.rate,
                                         spec_node, S_v, S_w, 0)
                    self.branches[array_id] = new_branch
                else:
                    array_id = len(self.branches)
                    new_branch = _Branch(self.id_counter, array_id, branch.rate,
                                         spec_node, S_v, S_w, 0)
                    self.branches.append(new_branch)
                
                self.ES_to_b[(S_v, S_w)].append(new_branch)
                self.id_counter += 1
                
            self.total_rate += (len(S_v.children) - 1) * branch.rate
            self.total_surviving += len(S_v.children) - 1
            
    
    def _duplication(self, event_tstamp, branch):
        
        S_u, S_v = branch.S_u, branch.S_v
        
        dupl_node = PhyloTreeNode(branch.ID, label='D',
                                  color=(S_u.ID,S_v.ID),
                                  tstamp=event_tstamp,
                                  dist=abs(event_tstamp-branch.parent.tstamp),
                                  transferred=branch.transferred)
        branch.parent.add_child(dupl_node)
        self.ES_to_b[(S_u, S_v)].remove(branch)
        
        copy_number = self._get_copy_number()
        
        for i in range(copy_number):
            
            if i == 0:
                array_id = branch.array_id
                new_branch = _Branch(self.id_counter, array_id, self.rate_sum,
                                     dupl_node, S_u, S_v, 0)
                self.branches[array_id] = new_branch
            else:
                array_id = len(self.branches)
                new_branch = _Branch(self.id_counter, array_id, self.rate_sum,
                                     dupl_node, S_u, S_v, 0)
                self.branches.append(new_branch)
            
            self.ES_to_b[(S_u, S_v)].append(new_branch)
            self.id_counter += 1
        
        self.total_surviving += (copy_number - 1)
        self.total_rate += (copy_number - 1) * (self.rate_sum)
        
        if (self._prohibit_extinction == 'per_species' and
            len(self.ES_to_b[(S_u, S_v)]) == copy_number):
            self.total_rate += self.l
            
            
    def _loss(self, event_tstamp, branch):
        
        S_u, S_v = branch.S_u, branch.S_v
        
        loss_node = PhyloTreeNode(branch.ID, label='*',
                                  color=(S_u.ID,S_v.ID),
                                  tstamp=event_tstamp,
                                  dist=abs(event_tstamp-branch.parent.tstamp),
                                  transferred=branch.transferred)
        branch.parent.add_child(loss_node)
        self.ES_to_b[(S_u, S_v)].remove(branch)
        branch.rate = 0.0
        
        self.total_surviving -= 1
        self.total_rate -= self.rate_sum
        
        if (self._prohibit_extinction == 'per_species' and
            len(self.ES_to_b[(S_u, S_v)]) == 1):
            self.total_rate -= self.l
            self.ES_to_b[(S_u, S_v)][0].rate -= self.l
            
    
    
    def _hgt(self, event_tstamp, branch):
        
        S_u, S_v = branch.S_u, branch.S_v
        
        valid_edges = self._coexisting_edges(event_tstamp,
                                             exclude_edge=(S_u, S_v))
        if valid_edges:
            trans_edge = random.choice(valid_edges)
            hgt_node = PhyloTreeNode(branch.ID, label='H',
                                     color=(S_u.ID,S_v.ID),
                                     tstamp=event_tstamp,
                                     dist=abs(event_tstamp-branch.parent.tstamp),
                                     transferred=branch.transferred)
            branch.parent.add_child(hgt_node)
            self.ES_to_b[(S_u, S_v)].remove(branch)
            
            # original branch
            array_id = branch.array_id
            new_branch = _Branch(self.id_counter, array_id, branch.rate,
                                 hgt_node, S_u, S_v, 0) 
            self.branches[array_id] = new_branch
            self.ES_to_b[(S_u, S_v)].append(new_branch)
            self.id_counter += 1
            
            # receiving branch
            array_id = len(self.branches)
            trans_branch = _Branch(self.id_counter, array_id, self.rate_sum,
                                   hgt_node, *trans_edge, 1)
            self.branches.append(trans_branch)
            self.ES_to_b[trans_edge].append(trans_branch)
            self.id_counter += 1
            
            self.total_surviving += 1
            self.total_rate += self.rate_sum
            
            if (self._prohibit_extinction == 'per_species' and
                len(self.ES_to_b[trans_edge]) == 2):
                self.total_rate += self.l
                self.ES_to_b[trans_edge][0].rate += self.l 
                
    
    def _assert_no_extinction(self, T):
        """Returns False if gene family is extinct in some species."""
        
        VS_to_VT = {l.ID: [] for l in self.S.preorder() if not l.children}
        
        for v in T.preorder():
            if not v.children and not v.is_loss():
                VS_to_VT[v.color].append(v.ID)
                
        for leaf_list in VS_to_VT.values():
            if not leaf_list:
                return False
        
        return True
        

# --------------------------------------------------------------------------
#                 CONSTRUCTION OF THE OBSERVABLE TREE
# --------------------------------------------------------------------------

def observable_tree(tree):
    
    obs_tree = delete_losses_and_contract(tree, inplace=False)
    
    remove_planted_root(obs_tree, inplace=True)
    
    return obs_tree


# -*- coding: utf-8 -*-

import os

import numpy as np

from asymmetree import PhyloTree
import asymmetree.treeevolve as te
import asymmetree.seqevolve as se
from asymmetree.file_io.SeqFileIO import write_alignment, write_fasta


__author__ = 'David Schaller'


class GenomeSimulator:
    
    def __init__(self, species_tree, outdir=None):
        
        if not isinstance(species_tree, PhyloTree):
            raise TypeError("species tree must be of type 'PhyloTree'")
        
        self.S = species_tree
        self.outdir = outdir
        
        if self.outdir:
            self._check_outdir()
            self.S.serialize(self._path('species_tree.json'), mode='json')
            
        self.true_gene_trees = []
        self.observable_gene_trees = []
        
    
    def _check_outdir(self):
        
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)
            
        elif os.path.exists(self.outdir) and not os.path.isdir(self.outdir):
            raise FileExistsError("'{}' is not a directory".format(self.outdir))
            
        for directory in ('true_gene_trees', 'fasta_files', 'alignments'):
            path = os.path.join(self.outdir, directory)
            if not os.path.exists(path):
                os.makedirs(path)
            
    
    def _path(self, *args):
        
        return os.path.join(self.outdir, *args)
            
    
    def simulate_gene_trees(self, N,
                            **kwargs):
        
        self.number_of_families = N
        
        self.true_gene_trees = te.simulate_gene_trees(self.S, N=N, **kwargs)
        if N == 1:
            self.true_gene_trees = [self.true_gene_trees]
        
        self.observable_gene_trees = [te.observable_tree(tree)
                                      for tree in self.true_gene_trees]
        
        # sequences should be emptied here if methods were called before
        if hasattr(self, 'sequence_dicts'):
            self.sequence_dicts.clear()
        
        if self.outdir:
            for i in range(N):
                filename = self._path('true_gene_trees',
                                      'gene_tree{}.json'.format(i))
                self.true_gene_trees[i].serialize(filename, mode='json')
                
    
    def simulate_sequences(self, subst_model,
                           root_genome=None,
                           length_distr=('constant', 200),
                           min_length=10,
                           write_fastas=True,
                           write_alignments=True,
                           **kwargs):
        
        self.subst_model = subst_model
        
        if hasattr(self, 'sequence_dicts'):
            self.sequence_dicts.clear()
        else:
            self.sequence_dicts = []
        
        if root_genome:
            if len(root_genome) != len(self.number_of_families):
                raise ValueError('no. of sequences in root genome does not'\
                                 'match no of gene families')
        else:
            self._check_distribution(length_distr, min_length)
        
        evolver = se.Evolver(subst_model, **kwargs)
        
        for i in range(self.number_of_families):
            
            OGT = self.observable_gene_trees[i]
            
            if root_genome:
                evolver.evolve_along_tree(OGT, start_seq=root_genome[i].upper())
                
            else:
                start_length = self._get_sequence_length()
                evolver.evolve_along_tree(OGT, start_length=start_length)
                
            self.sequence_dicts.append(evolver.sequences)
            
            # write one alignment file per gene family
            if self.outdir and write_alignments:
                self._write_alignment(i)
        
        # write one fasta file per species
        if self.outdir and write_fastas:
            self._write_fastas(include_inner=False)
            
            
    def _write_alignment(self, family_id):
        
        if not self.outdir:
            raise RuntimeError('no output directory specified for alignments')        
        
        alg_builder = se.AlignmentBuilder(self.observable_gene_trees[family_id],
                                          self.sequence_dicts[family_id],
                                          self.subst_model.alphabet,
                                          include_inner=False)
        
        alignment = []
        for node, sequence in alg_builder.build().items():
            label = self._compose_label(node, family_id)
            alignment.append( (label, sequence) )
            
        basename = 'alignment{}.phylip'.format(family_id)
        filename = self._path('alignments', basename)
        write_alignment(filename, alignment, alignment_format='phylip')
    
                
    def _write_fastas(self, include_inner=False):
        
        if not self.outdir:
            raise RuntimeError('no output directory specified for fasta files')
        
        # list of all species IDs
        if include_inner:
            species = [v.ID for v in self.S.preorder()]
        else:
            species = [v.ID for v in self.S.preorder() if not v.children]
        
        # labeled sequences sorted by color/species
        sorted_seqs = {s: [] for s in species}
        
        for i in range(self.number_of_families):
            for node, evoseq in self.sequence_dicts[i].items():
                
                # skip inner nodes
                if not include_inner and node.children:
                    continue
                
                label = self._compose_label(node, i)
                sequence = self.subst_model.to_sequence(evoseq)
                sorted_seqs[node.color].append( (label, sequence) )
                
        for spec, sequences in sorted_seqs.items():
            basename = '{}.f{}a'.format(spec, self.subst_model.model_type)
            filename = self._path('fasta_files', basename)
            write_fasta(filename, sequences)
                
                
    def _check_distribution(self, distr, min_length):
        
        if not isinstance(min_length, int) or min_length < 0:
            raise ValueError('minimal sequence length must be an int >=0')
        self._min_length = min_length
        
        if isinstance(distr, int):
            if min_length > distr:
                raise ValueError('constant sequence length must be >= '\
                                 'minimal sequence length')
                
            self._length_distr = 'constant'
            self._seq_length = distr
            
        elif distr[0] == 'constant':
            seq_length = distr[1]
            if not isinstance(seq_length, int) or min_length > seq_length:
                raise ValueError('constant sequence length must be an int >= '\
                                 'minimal sequence length')
                
            self._length_distr = 'constant'
            self._seq_length = seq_length
            
        elif distr[0] == 'gamma':
            shape = distr[1]
            scale = distr[2]
            
            if (not isinstance(shape, float) or shape <= 0.0 or 
                not isinstance(scale, float) or scale <= 0.0):
                raise ValueError('scale and shape parameters for gamma '\
                                 'distribution must be floats >0.0')
            
            mean = round(shape * scale, 3)
            if mean < min_length:
                raise ValueError('expected value for gamma distribution '\
                                 '({}) must be >= minimal length'.format(mean))
            
            self._length_distr == 'gamma'
            self._shape = shape
            self._scale = scale
            
        elif distr[0] == 'gamma_mean':
            mean = distr[1]
            
            if not isinstance(mean , (int, float)) or mean < min_length:
                raise ValueError('mean of gamma distribution must be '\
                                 '>= minimal length')
            
            self._length_distr == 'gamma'
            self._shape = 1.0
            self._scale = mean / self._shape
            
        else:
            raise ValueError("length distribution '{}' is not "\
                             "supported".format(distr))
            
    
    def _get_sequence_length(self):
        
        if self._length_distr == 'constant':
            return self._seq_length
        elif self._length_distr == 'gamma':
            length = -1
            while length < self._min_length:
                length = round(np.random.gamma(self._shape, scale=self._scale))
            return length
        
    
    def _compose_label(self, node, family_id):
        
        return 'fam{}gene{}spec{}'.format(family_id, node.label, node.color)